# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Thorsten Simons (sw@snomis.de)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import logging
import socket
from http.client import HTTPConnection as _HTTPConnection, HTTPS_PORT

__all__ = ['HTTPConnection']

# from netinet/tcp.h:
#   define TCP_KEEPALIVE   0x10    /* idle time used when SO_KEEPALIVE is enabled */
#   The TCP_KEEPALIVE options enable to specify the amount of time, in
#   seconds, that the connection must be idle before keepalive probes
#   (if enabled) are sent.  The default value is specified by the MIB
#   variable net.inet.tcp.keepidle.
TCP_KEEPALIVE = 0x10
# from netinet/tcp.h:
#   define	TCP_KEEPINTVL	0x101	/* interval between keepalives */
#   When keepalive probes are enabled, this option will set the amount
#   of time in seconds between successive keepalives sent to probe an
#   unresponsive peer.
TCP_KEEPINTVL = 0x101
# from netinet/tcp.h:
#   define	TCP_KEEPCNT		0x102	/* number of keepalives before close */
#   When keepalive probes are enabled, this option will set the number
#   of times a keepalive probe should be repeated if the peer is not
#   responding. After this many probes, the connection will be closed.
TCP_KEEPCNT = 0x102

logging.getLogger('hcpsdk.httpclient').addHandler(logging.NullHandler())


class HTTPConnection(_HTTPConnection):
    """
    Subclass of http.client.HTTPConnection that allows for TCP keep-alive.

    This copies the *__init__()* and *connect()* methods, adding in the
    necessary code to enable TCP keep-alive.
    """

    def __init__(self, host, port=None, timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
                 source_address=None, sock_keepalive=False,
                 tcp_keepalive=60, tcp_keepintvl=60, tcp_keepcnt=3):
        """
        :param host:            target host (fqdn or ip-address
        :param port:            target port
        :param timeout:         blocking calls time-out
        :param source_address:  source address tuple (ip. port) to be used
        :param sock_keepalive:  enable TCP keepalive, if True
        :param tcp_keepalive:   idle time used when SO_KEEPALIVE is enable
        :param tcp_keepintvl:   interval between keepalives
        :param tcp_keepcnt:     number of keepalives before close
        """

        self.logger = logging.getLogger(__name__ + '.HTTPConnection')
        self.sock_keepalive = sock_keepalive
        self.tcp_keepalive = tcp_keepalive
        self.tcp_keepintvl = tcp_keepintvl
        self.tcp_keepcnt = tcp_keepcnt
        super().__init__(host, port, timeout=timeout,
                         source_address=source_address)

    def connect(self):
        """
        Connect to the host and port specified in __init__, using the
        keep-alive settings specified there as well.
        """
        self.sock = self._create_connection(
            (self.host,self.port), self.timeout, self.source_address)

        # added to standard method:
        if self.sock_keepalive:
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            self.sock.setsockopt(socket.SOL_TCP, TCP_KEEPALIVE, self.tcp_keepalive)
            self.sock.setsockopt(socket.SOL_TCP, TCP_KEEPINTVL, self.tcp_keepintvl)
            self.sock.setsockopt(socket.SOL_TCP, TCP_KEEPCNT, self.tcp_keepcnt)

            # self.logger.debug('enabled TCP keep-alive (TCP_KEEPALIVE = {}, '
            print('enabled TCP keep-alive (TCP_KEEPALIVE = {}, '
                              'TCP_KEEPINTVL = {}, TCP_KEEPCNT = {})'
                              .format(self.sock.getsockopt(socket.SOL_TCP,
                                                           TCP_KEEPALIVE),
                                      self.sock.getsockopt(socket.SOL_TCP,
                                                           TCP_KEEPINTVL),
                                      self.sock.getsockopt(socket.SOL_TCP,
                                                           TCP_KEEPCNT)))
        # end of addition

        self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        if self._tunnel_host:
            self._tunnel()


try:
    import ssl
except ImportError:
    pass
else:
    class HTTPSConnection(_HTTPConnection):
        "This class allows communication via SSL."

        default_port = HTTPS_PORT

        # XXX Should key_file and cert_file be deprecated in favour of context?

        def __init__(self, host, port=None, key_file=None, cert_file=None,
                     timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
                     source_address=None, *, context=None,
                     check_hostname=None, sock_keepalive=False,
                     tcp_keepalive=60, tcp_keepintvl=60, tcp_keepcnt=3):
            """
            :param host:            target host (fqdn or ip-address
            :param port:            target port
            :param timeout:         blocking calls time-out
            :param source_address:  source address tuple (ip. port) to be used
            :param sock_keepalive:  enable TCP keepalive, if True
            :param tcp_keepalive:   idle time used when SO_KEEPALIVE is enable
            :param tcp_keepintvl:   interval between keepalives
            :param tcp_keepcnt:     number of keepalives before close
            """
            # added to standard method:
            self.logger = logging.getLogger(__name__ + '.HTTPConnection')
            self.sock_keepalive = sock_keepalive
            self.tcp_keepalive = tcp_keepalive
            self.tcp_keepintvl = tcp_keepintvl
            self.tcp_keepcnt = tcp_keepcnt
            # end of addition

            super(HTTPSConnection, self).__init__(host, port, timeout,
                                                  source_address)
            self.key_file = key_file
            self.cert_file = cert_file
            if context is None:
                context = ssl._create_default_https_context()
            will_verify = context.verify_mode != ssl.CERT_NONE
            if check_hostname is None:
                check_hostname = context.check_hostname
            if check_hostname and not will_verify:
                raise ValueError("check_hostname needs a SSL context with "
                                 "either CERT_OPTIONAL or CERT_REQUIRED")
            if key_file or cert_file:
                context.load_cert_chain(cert_file, key_file)
            self._context = context
            self._check_hostname = check_hostname

        def connect(self):
            "Connect to a host on a given (SSL) port."

            super().connect()

            # added to standard method:
            if self.sock_keepalive:
                self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                self.sock.setsockopt(socket.SOL_TCP, TCP_KEEPALIVE, self.tcp_keepalive)
                self.sock.setsockopt(socket.SOL_TCP, TCP_KEEPINTVL, self.tcp_keepintvl)
                self.sock.setsockopt(socket.SOL_TCP, TCP_KEEPCNT, self.tcp_keepcnt)

                # self.logger.debug('enabled TCP keep-alive (TCP_KEEPALIVE = {}, '
                print('enabled TCP keep-alive (TCP_KEEPALIVE = {}, '
                                  'TCP_KEEPINTVL = {}, TCP_KEEPCNT = {})'
                                  .format(self.sock.getsockopt(socket.SOL_TCP,
                                                               TCP_KEEPALIVE),
                                          self.sock.getsockopt(socket.SOL_TCP,
                                                               TCP_KEEPINTVL),
                                          self.sock.getsockopt(socket.SOL_TCP,
                                                               TCP_KEEPCNT)))
            # end of addition

            if self._tunnel_host:
                server_hostname = self._tunnel_host
            else:
                server_hostname = self.host

            self.sock = self._context.wrap_socket(self.sock,
                                                  server_hostname=server_hostname)
            if not self._context.check_hostname and self._check_hostname:
                try:
                    ssl.match_hostname(self.sock.getpeercert(), server_hostname)
                except Exception:
                    self.sock.shutdown(socket.SHUT_RDWR)
                    self.sock.close()
                    raise

    __all__.append("HTTPSConnection")



import sys
from os.path import join, normpath
import logging
import threading
import time
import pprint
import hcpsdk

T_NAMESPACE = 'n1.m.hcp73.archivas.com'
T_DAAC      = 'n'
T_DAAC_PWD  = 'n01'
T_PORT      = 443

T_THREADS        = 10          # no. of threads to use in parallel
T_OBJ            = 128            # the object we will use in a run (from objs)
T_CNT_PER_THREAD = 100          # how often the object will be written
T_OUTNAME   = '_hcp_loadtest'   # filename to be used
T_STARTPATH = '/rest/_hcp_loadtest/'    # path to write to

# a names of the source objects
T_OBJS = {     1: '1kbfile',
               8: '8kbfile',
              16: '16kbfile',
              32: '32kbfile',
              64: '64kbfile',
             128: '128kbfile',
             256: '256kbfile',
             512: '512kbfile',
            1024: '1mbfile',
           10240: '10mbfile',
          102400: '100mbfile'}



if __name__ == '__main__':
    strhandler = logging.StreamHandler(sys.stdout)
    frm = logging.Formatter("\t%(asctime)s [%(levelname)-8s] %(threadName)s:%(name)s.%(funcName)s.%(lineno)d: %(message)s") #, "%m/%d %H:%M:%S")
    l = logging.getLogger()
    strhandler.setFormatter(frm)
    l.addHandler(strhandler)
#    l.setLevel(logging.WARNING)
    l.setLevel(logging.DEBUG)

    # the place where we will find the needed source objects
    if sys.platform == 'win32':
        T_INPATH = 'd:\\__files'
    elif sys.platform == 'darwin':
        T_INPATH = '/Volumes/dev/__files'
    else:
        sys.exit('source path is undefined')

    # the paths in which we will store the objects
    i1 = normpath(join(T_STARTPATH, T_OUTNAME, '1')).replace('\\', '/')
    i2 = normpath(join(T_STARTPATH, T_OUTNAME, '2')).replace('\\', '/')

    # load one of the files into memory
    print('--> Loading test object into memory - ', end='')
    with open(join(T_INPATH,T_OBJS[T_OBJ]), 'rb') as inHdl:
        buf = bytearray(inHdl.read())
        size = len(buf)
        print('object size = {}'.format(size))

    print("--> Init <hcpsdk.target> object")
    try:
        hcptarget = hcpsdk.target(T_NAMESPACE, T_DAAC, T_DAAC_PWD, T_PORT)
    except hcpsdk.HcpsdkError as e:
        sys.exit("Fatal: {}".format(e.errText))

    print("--> Init <hcpsdk.connection> object")
    try:
        con = hcpsdk.connection(hcptarget, idletime=3, debuglevel=9)
    except Exception as e:
        sys.exit('Exception: {}'.format(str(e)))

    print("--> Write 1st file:", i1)
    try:
        r = con.PUT(i1, buf)
        print(r.status)
    except Exception as e:
        print('Exception: {}'.format(str(e)))

    secs = 10
    print("--> sleep for {}".format(secs))
    time.sleep(secs)

    print("--> Write 2nd file:", i2)
    try:
        r = con.PUT(i2, buf)
        print(r.status)
    except Exception as e:
        print('Exception: {}'.format(str(e)))

    print("--> HEAD 1st file:", i1)
    try:
        r = con.HEAD(i1)
        print(r.status)
    except Exception as e:
        print('Exception: {}'.format(str(e)))

    print("--> HEAD 2nd file:", i2)
    try:
        r = con.HEAD(i2)
        print(r.status)
    except Exception as e:
        print('Exception: {}'.format(str(e)))

    print("--> Set indexing on 1st file:", i1)
    try:
        r = con.POST(i1, ['index=true'])
        print(r.status)
    except Exception as e:
        print('Exception: {}'.format(str(e)))

    print("--> Set indexing on 2nd file:", i2)
    try:
        r = con.POST(i2, ['index=true'])
        print(r.status)
    except Exception as e:
        print('Exception: {}'.format(str(e)))

    print("--> Read 1st file:", i1)
    try:
        r = con.GET(i1)
        print(r.status)
    except Exception as e:
        print('Exception: {}'.format(str(e)))
    else:
        con.read()

    print("--> Read 2nd file:", i2)
    try:
        r = con.GET(i2)
        print(r.status)
    except Exception as e:
        print('Exception: {}'.format(str(e)))
    else:
        con.read()

    print("--> Delete 1st file:", i1)
    try:
        r = con.DELETE(i1)
        print(r.status)
    except Exception as e:
        print('Exception: {}'.format(str(e)))

    print("--> Delete 2nd file:", i2)
    try:
        r = con.DELETE(i2)
        print(r.status)
    except Exception as e:
        print('Exception: {}'.format(str(e)))

    print("--> Close <hcpsdk.connection> object")
    con.close()

    while True:
        ts = threading.active_count()
        l.info('--> Threads: {}'.format(ts))
        if ts == 1:
            break
        time.sleep(1)



# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Thorsten Simons (sw@snomis.de)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import time

from datetime import date
import xml.etree.ElementTree as Et
from collections import OrderedDict
from tempfile import TemporaryFile, NamedTemporaryFile
import logging
import hcpsdk


__all__ = ['LogsError', 'LogsNotReadyError', 'LogsInProgessError', 'Logs']

logging.getLogger('hcpsdk.mapi.logs').addHandler(logging.NullHandler())


class LogsError(Exception):
    """
    Base Exception used by the *hcpsdk.mapi.Logs()* class.
    """
    def __init__(self, reason):
        """
        :param reason: An error description
        """
        self.args = (reason,)

class LogsNotReadyError(LogsError):
    """
    Raised by *Logs.download()* in case there are no logs ready to be
    downloaded.
    """
    def __init__(self, reason):
        """
        :param reason: An error description
        """
        self.args = (reason,)

class LogsInProgessError(LogsError):
    """
    Raised by *Logs.download()* in case the log download is already
    in progress.
    """
    def __init__(self, reason):
        """
        :param reason: An error description
        """
        self.args = (reason,)


class Logs(object):
    """
    Access to HCP internal logfiles (ACCESS, SYSTEM, SERVICE, APPLICATION)
    """

    L_ACCESS = 'ACCESS' # http access logs
    L_SYSTEM = 'SYSTEM' #
    L_SERVICE = 'SERVICE' #
    L_APPLICATION = 'APPLICATION' #
    L_ALL = [L_ACCESS, L_SYSTEM, L_SERVICE, L_APPLICATION]

    def __init__(self, target, debuglevel=0):
        """
        :param target:      an hcpsdk.Target object
        :param debuglevel:  0..9 (used in *http.client*)
        :raises:            *hcpsdk.HcpsdkPortError* in case *target* is
                            initialized with an incorrect port for use by
                            this class.
        """
        self.logger = logging.getLogger(__name__ + '.Logs')
        hcpsdk.checkport(target, hcpsdk.P_MAPI)
        self.target = target
        self.debuglevel = debuglevel
        self.connect_time = 0.0
        self.service_time = 0.0
        self.prepare_xml = None
        self.suggestedfilename = '' # the filename suggested by HCP

        try:
            self.con = hcpsdk.Connection(self.target, debuglevel=self.debuglevel)
        except Exception as e:
            raise hcpsdk.HcpsdkError(str(e))

    def mark(self, message):
        """
        Mark HCPs internal log with a message.

        :param message: the message to be logged
        :raises:        *LogsError*
        """
        if not message:
            raise LogsError('log message required')

        self.logger.debug('mark log with "{}"'.format(message))
        try:
            self.con.POST('/mapi/logs', params={'mark': message})
        except Exception as e:
            self.logger.error(e)
            raise LogsError(e)
        else:
            self.con.read()
            if self.con.response_status != 200:
                raise LogsError('{} - {} ({})'
                                .format(self.con.response_status,
                                        self.con.response_reason,
                                        self.con.getheader('X-HCP-ErrorMessage',
                                                                    default='?')))

    def prepare(self, startdate=None, enddate=None, snodes=[]):
        """
        Command HCP to prepare logs from *startdate* to *enddate* for
        later download.

        :param startdate:   1st day to collect (as a *datetime.date* object)
        :param enddate:     last day to collect (as a *datetime.date* object)
        :param snodes:      list of S-series nodes to collect from
        :returns:           a tuple of datetime.date(startdate),
                            datetime.date(enddate) and
                            str(prepared XML)
        :raises:            *ValueError* arguments are invalid or one of the
                            *LogsError* if an operation failed
        """
        if startdate and type(startdate) != date:
            raise ValueError('startdate not of type(datetime.date)')
        else:
            self.startdate = startdate or date(1970,1,1)
        if enddate and type(enddate) != date:
            raise ValueError('enddate not of type(datetime.date)')
        else:
            self.enddate = enddate or date.today()
        if type(snodes) != list:
            raise ValueError('snodes not of type(list)')

        self.logger.debug('preparing logs for {} to {}'
                          .format(self.startdate.isoformat(),
                                  self.enddate.isoformat()))

        # Prepare the XML command file
        if snodes:
            xsnodes = '<snodes>' + ','.join(snodes) + '</snodes>'
        else:
            xsnodes = ''
        self.prepare_xml = '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n' \
                           '<logPrepare>\n' \
                           '  <startDate>{:02}/{:02}/{:04}</startDate>\n' \
                           '  <endDate>{:02}/{:02}/{:04}</endDate>\n' \
                           '  {}\n' \
                           '</logPrepare>\n'\
                            .format(self.startdate.month, self.startdate.day,
                                    self.startdate.year, self.enddate.month,
                                    self.enddate.day, self.enddate.year,
                                    xsnodes)
        self.logger.debug('prepare_xml: {}'.format(self.prepare_xml))

        try:
            self.con.POST('/mapi/logs/prepare', body=self.prepare_xml,
                          headers={'Content-Type': 'application/xml'})
        except Exception as e:
            self.logger.error(e)
            raise LogsError(e)
        else:
            self.logger.debug('result: {} - {}'.format(self.con.response_status,
                                                       self.con.response_reason))
            self.logger.debug('returned headers: {}'.format(self.con.getheaders()))

            self.con.read()
            if self.con.response_status == 200:
                return(self.startdate, self.enddate, self.prepare_xml)
            elif self.con.response_status == 400:
                raise LogsInProgessError('{} - {} ({})'
                                         .format(self.con.response_status,
                                                 self.con.response_reason,
                                                 self.con.getheader('X-HCP-ErrorMessage',
                                                                    default='?')))
            else:
                raise LogsError('{} - {} ({})'
                                .format(self.con.response_status,
                                        self.con.response_reason,
                                        self.con.getheader('X-HCP-ErrorMessage',
                                                           default='?')))

    def status(self):
        """
        Query HCP for the status of the request log download.

        :returns:   a *collection.OrderedDict*:
                    ::

                        {
                         readyForStreaming: bool,
                         streamingInProgress: bool,
                         started: bool,
                         error: bool,
                         content: list # one or more of: L_ACCESS, L_SYSTEM,
                                       # L_SERVICE, L_APPLICATION)
                        }
        :raises:    re-raises whatever is raised below
        """
        self.logger.debug('status query issued')

        try:
            self.con.GET('/mapi/logs')
        except Exception as e:
            self.logger.error(e)
            raise
        else:
            self.logger.debug('response headers: {}'.format(self.con.getheaders()))
            xml = self.con.read().decode()
            time.sleep(.5)

            if self.con.response_status != 200:
                return None
            else:
                stat = OrderedDict()
                for child in Et.fromstringlist(xml):
                    if child.text == 'true':
                        stat[child.tag] = True
                    elif child.text == 'false':
                        stat[child.tag] = False
                    else:
                        stat[child.tag] = child.text.split(',')
                return stat

    def download(self, hdl=None, nodes=[], snodes=[], logs=[],
                 progresshook=None, hidden=True):
        """
        Download the requested logs.

        :param hdl:     a file (or file-like) handle open for binary
                        read/write or *None*, in which case a temporary file
                        will be created
        :param nodes:   list of node-IDs (int), all if empty
        :param snodes:  list of S-node names (str), none if empty
        :param logs:    list of logs (*L_**), all if empty
        :param progresshook:    a function taking a single argument (the #
                                of bytes received) that will be called after
                                each chunk of bytes downloaded
        :param hidden:  the temporary file created will be hidden if possible
                        (see `tempfile.TemporaryFile()
                        <https://docs.python.org/3/library/tempfile.html#tempfile.TemporaryFile>`_)
        :returns:       a 2-tuple: the file handle holding the received logs,
                        positioned at byte 0 and the filename suggested by
                        HCP
        :raises:        *LogsError* or *LogsNotReadyError*
        """

        # make sure we have a file handle open for binary read/write
        if not hdl:
            if hidden:
                self.hdl = TemporaryFile('w+b')
            else:
                self.hdl = NamedTemporaryFile('w+b')
        else:
            self.hdl = hdl

        # create the XML command structire
        str_nodes = str_snodes = str_logs = ''
        if nodes:
            str_nodes = '<nodes>' + ','.join(nodes) + '</nodes>'
        else:
            str_nodes = ''
        if snodes:
            str_snodes = ','.join(snodes)
        if logs:
            str_logs = ','.join(logs)
        else:
            str_logs = ','.join(Logs.L_ALL)

        xml = '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n' \
              '<logDownload>\n' \
              '    {}\n' \
              '    <snodes>{}</snodes>\n' \
              '    <content>{}</content>\n' \
              '</logDownload>'.format(str_nodes, str_snodes, str_logs).encode()

        self.logger.debug('dl_xml: {}'.format(xml))

        # download the logs
        try:
            self.con.POST('/mapi/logs/download', body=xml,
                          headers={'Accept': '*/*',
                                   'Content-Type': 'application/xml'})
        except Exception as e:
            self.logger.error(e)
            raise LogsError(e)
        else:
            self.logger.debug('result: {} - {}'.format(self.con.response_status,
                                                       self.con.response_reason))
            self.logger.debug('returned headers: {}'.format(self.con.getheaders()))
            suggestedfilename = self.con.getheader('Content-Disposition',
                                                   'name=no-name').split('=')[1]

        if self.con.response_status == 200:
            numbytes = 0
            try:
                while True:
                    d = self.con.read(amt=2**18)
                    numbytes += len(d)
                    if progresshook:
                        progresshook(numbytes)
                    if d:
                        self.hdl.write(d)
                    else:
                        break
            except Exception as e:
                raise LogsError(e)
        else:
            try:
                self.con.read()
            except Exception as e:
                raise LogsError(e)
            raise LogsError('{} - {} ({})'.format(self.con.response_status,
                                                  self.con.response_reason,
                                                  self.con.getheader('X-HCP-ErrorMessage',
                                                                     './.')))
        self.hdl.seek(0)
        return (self.hdl, suggestedfilename)

    def cancel(self):
        """
        Cancel a log request.

        :returns:   *True* if cancel was successfull
        :raises:    *LogsError* in case the cancel failed
        """
        self.logger.debug('cancel request issued')

        try:
            self.con.POST('/mapi/logs', params={'cancel': ''})
        except Exception as e:
            self.logger.error(e)
            raise LogsError(e)
        else:
            self.con.read() # cleanup

        if self.con.response_status == 200:
            self.suggestedfilename = ''
            return True
        else:
            self.logger.debug('{} - {} (cancel failed)'
                              .format(self.con.response_status,
                                    self.con.response_reason))
            raise LogsError('{} - {} (cancel failed)'
                            .format(self.con.response_status,
                                    self.con.response_reason))

    def close(self):
        """
        Close the underlying *hcpsdk.Connection()*.
        """
        self.logger.debug('close Logs()')
        self.suggestedfilename = ''
        self.con.close()





# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Thorsten Simons (sw@snomis.de)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import unittest

from hcpsdk.pathbuilder import PathBuilder


class TestHcpsdk_50_1_PathBuilder(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_1_10_buildpath(self):
        """
        Make sure we get a tuple containing a path and a UUID as string
        """
        print('test_1_10_buildpath:')
        b = PathBuilder()
        t = b.getunique('testfile.txt')
        print(t)
        self.assertTrue(type(t) == tuple)
        self.assertTrue(len(t) == 2)

    def test_1_20_buildpath_annotation(self):
        """
        Make sure we get a tuple containing a path and a UUID as string
        """
        print('test_1_20_buildpath_annotation:')
        b = PathBuilder(annotation=True)
        t = b.getunique('testfile.txt')
        print(t[0:2])
        for i in t[2].split(sep='\n'):
            print(i)
        self.assertTrue(type(t) == tuple)
        self.assertTrue(len(t) == 3)


if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Thorsten Simons (sw@snomis.de)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import hcpsdk
import xml.etree.ElementTree as Et
from collections import OrderedDict
import logging

__all__ = ['Info']

logging.getLogger('hcpsdk.namespace').addHandler(logging.NullHandler())


class Info(object):
    """
    Class to access namespaces metadata information
    """

    def __init__(self, target, debuglevel=0):
        """
        :param target:      an **hcpsdk.Target** object
        :param debuglevel:  0..9 (propagated to *http.client*)
        """
        self.logger = logging.getLogger(__name__ + '.Info')
        self.target = target
        self.debuglevel = debuglevel
        self.connect_time = 0.0
        self.service_time = 0.0

    def nsstatistics(self):
        """
        Query for namespace statistic information

        :return:  a dict holding the stats
        :raises: hcpsdk.HcpsdkError()
        """
        # noinspection PyUnusedLocal
        d = None
        try:
            con = hcpsdk.Connection(self.target, debuglevel=self.debuglevel)
        except Exception as e:
            raise hcpsdk.HcpsdkError(str(e))
        else:
            self.connect_time = con.connect_time
            try:
                r = con.GET('/proc/statistics')
            except Exception as e:
                raise hcpsdk.HcpsdkError(str(e))
            else:
                if r.status == 200:
                    # Good status, get and parse the Response
                    x = r.read()
                    self.service_time = con.service_time2
                    root = Et.fromstring(x)
                    d = root.attrib
                    tobedel = None
                    for i in d.keys():
                        if i.startswith('{http'):
                            tobedel = i
                        else:
                            d[i] = self._castvar(d[i])
                    if tobedel:
                        del d[tobedel]
                else:
                    raise (hcpsdk.HcpsdkError('{} - {}'.format(r.status, r.reason)))
        finally:
            # noinspection PyUnboundLocalVariable
            con.close()

        return d

    # noinspection PyShadowingBuiltins
    def listaccessiblens(self, all=False):
        """
        List the settings of the actual (or all accessible namespace(s).

        :param all:     list all accessible namespaces if True, else list the
                        actual one, only.
        :return:        a dict holding a dict per namespace
        """

        # setup Target URL and apply parameters
        url = '/proc'
        if not all:
            params = {'single': 'true'}
        else:
            params = None

        try:
            con = hcpsdk.Connection(self.target, debuglevel=self.debuglevel)
        except Exception as e:
            raise hcpsdk.HcpsdkError(str(e))
        else:
            self.connect_time = con.connect_time

            try:
                r = con.GET(url, params=params)
            except Exception as e:
                raise hcpsdk.HcpsdkError(str(e))
            else:
                if r.status == 200:
                    # Good status, get and parse the Response
                    x = r.read()
                    self.service_time = con.service_time2
                    root = Et.fromstring(x)
                    d = OrderedDict()
                    for n in root:
                        d[n.attrib.get('name')] = n.attrib
                        for i in d[n.attrib.get('name')].keys():
                            d[n.attrib.get('name')][i] = \
                                self._castvar(d[n.attrib.get('name')][i])
                        for n1 in n:
                            d[n.attrib['name']]['description'] = n1.text.strip().split('°')
                else:
                    raise (hcpsdk.HcpsdkError('{} - {}'.format(r.status, r.reason)))
        finally:
            # noinspection PyUnboundLocalVariable
            con.close()

        return d

    def listretentionclasses(self):
        """
        List the Retention Classes available for the actual namespace.

        :return: a dict holding a dict per Retention Class
        """
        try:
            con = hcpsdk.Connection(self.target, debuglevel=self.debuglevel)
        except Exception as e:
            raise hcpsdk.HcpsdkError(str(e))
        else:
            self.connect_time = con.connect_time

            r = con.GET('/proc/retentionClasses')
            if r.status == 200:
                # Good status, get and parse the Response
                x = r.read()
                self.service_time = con.service_time2
                root = Et.fromstring(x)
                d = OrderedDict()
                for n in root:
                    d[n.attrib.get('name')] = n.attrib
                    for i in d[n.attrib.get('name')].keys():
                        d[n.attrib.get('name')][i] = \
                            self._castvar(d[n.attrib.get('name')][i])
                    for n1 in n:
                        d[n.attrib.get('name')]['description'] = n1.text.strip()
            else:
                raise (hcpsdk.HcpsdkError('{} - {}'.format(r.status, r.reason)))
        finally:
            # noinspection PyUnboundLocalVariable
            con.close()

        return d

    def listpermissions(self):
        """
        List the namespace and user permissions for the actual namespace.

        :return: a dict holding a dict per permission domain
        """
        try:
            con = hcpsdk.Connection(self.target, debuglevel=self.debuglevel)
        except Exception as e:
            raise hcpsdk.HcpsdkError(str(e))
        else:
            self.connect_time = con.connect_time

            r = con.GET('/proc/permissions')
            if r.status == 200:
                # Good status, get and parse the Response
                x = r.read()
                self.service_time = con.service_time2
                root = Et.fromstring(x)
                d = OrderedDict()
                for n in root:
                    d[n.tag] = n.attrib
                    for i in d[n.tag].keys():
                        d[n.tag][i] = self._castvar(d[n.tag][i])
            else:
                raise (hcpsdk.HcpsdkError('{} - {}'.format(r.status, r.reason)))
        finally:
            # noinspection PyUnboundLocalVariable
            con.close()

        return d

    # noinspection PyMethodMayBeStatic
    def _castvar(self, var):
        """
        Cast a value to the right type.
        :param var: a string
        :return: the casted value
        """
        if var == 'true':
            return True
        elif var == 'false':
            return False
        else:
            try:
                return int(var)
            except ValueError:
                return var



# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Thorsten Simons (sw@snomis.de)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from datetime import datetime, timedelta
from time import strftime
from io import StringIO
import logging
import hcpsdk


__all__ = ['ChargebackError', 'Chargeback']

logging.getLogger('hcpsdk.mapi.chargeback').addHandler(logging.NullHandler())

class ChargebackError(Exception):
    """
    Base Exception used by the *hcpsdk.mapi.Chargeback()* class.
    """
    def __init__(self, reason):
        """
        :param reason: An error description
        """
        self.args = (reason,)


class Chargeback(object):
    '''
    Access to HCP chargeback reports
    '''

    # the granularity modes allowed for chargeback collection
    CBG_DAY = 'day'
    CBG_HOUR = 'hour'
    CBG_TOTAL = 'total'
    CBG_ALL = [CBG_DAY, CBG_HOUR, CBG_TOTAL]

    # the output formats available
    CBM_CSV = 'text/csv'
    CBM_JSON = 'application/json'
    CBM_XML = 'application/xml'
    CBM_ALL = [CBM_CSV, CBM_JSON, CBM_XML]

    def __init__(self, target, timeout=600, debuglevel=0):
        '''
        :param target:      an hcpsdk.Target object pointing to an HCP FQDN
                            starting with **admin.** for access from a system
                            level account or **<tenant>.** for a tenant level
                            account
        :param timeout:     the connection timeout; relatively high per
                            default, as generating the report can take longer
                            than **hcpsdk**\ s default of 30 seconds on a busy
                            system
        :param debuglevel:  0..9 (used in *http.client*)
        '''
        self.logger = logging.getLogger(__name__ + '.Chargeback')
        hcpsdk.checkport(target, hcpsdk.P_MAPI)
        self.connect_time = 0.0
        self.service_time = 0.0

        try:
            self.con = hcpsdk.Connection(target, timeout=timeout,
                                         debuglevel=debuglevel)
        except Exception as e:
            raise hcpsdk.HcpsdkError(str(e))


    def request(self, tenant=None, start=None, end=None,
                granularity=CBG_TOTAL, fmt=CBM_JSON):
        '''
        Request a chargeback report for a Tenant.

        :param tenant:      the *Tenant* to collect from
        :param start:       starttime (a datetime object)
        :param end:         endtime (a datetime object)
        :param granularity: one out of CBG_ALL
        :param fmt:         output format, one out of CBM_ALL
        :return:            a file-like object in text-mode containing the
                            report
        '''
        if not tenant:
            raise ValueError('no tenant given')
        else:
            self.tenant = tenant
        if start and type(start) != datetime:
            raise ValueError('start not of type(datetime.datetime)')
        else:
            self.start = start or datetime.now() - timedelta(days=180)
        if end and type(end) != datetime:
            raise ValueError('end not of type(datetime.datetime)')
        else:
            self.end = end or datetime.now()
        if granularity not in Chargeback.CBG_ALL:
            raise ValueError('granularity not in {}'
                             .format(Chargeback.CBG_ALL))
        else:
            self.granularity = granularity
        if fmt not in Chargeback.CBM_ALL:
            raise ValueError('fmt not in {}'.format(Chargeback.CBM_ALL))
        else:
            self.fmt = fmt

        self.logger.debug('request for {} ({} - {}), {}, {}'
                          .format(self.tenant,
                                  self.start.isoformat(),
                                  self.end.isoformat(),
                                  self.granularity, self.fmt))

        try:
            self.con.GET('/mapi/tenants/{}/chargebackReport'.format(self.tenant),
                         params={'start': self.start.strftime('%Y-%m-%dT%H:%M:%S')+strftime('%z'),
                                 'end': self.end.strftime('%Y-%m-%dT%H:%M:%S')+strftime('%z'),
                                 'granularity': self.granularity,
                                 'prettyprint': 'true'},
                         headers={'Accept': self.fmt})
        except Exception as e:
            self.logger.error(e)
            raise ChargebackError(e)
        else:
            self.logger.debug('result: {} - {}'.format(self.con.response_status,
                                                       self.con.response_reason))
            self.logger.debug('returned headers: {}'.format(self.con.getheaders()))

            if self.con.response_status == 200:
                ret = StringIO(initial_value=self.con.read().decode())
                ret.seek(0)
                return ret
            else:
                # session cleanup!
                self.con.read()
                raise ChargebackError('{} - {} ({})'
                                      .format(self.con.response_status,
                                              self.con.response_reason,
                                              self.con.getheader('X-HCP-ErrorMessage',
                                                                 default='?')))

    def close(self):
        '''
        Close the underlying *hcpsdk.Connection* object.
        '''
        self.con.close()



# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Thorsten Simons (sw@snomis.de)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import xml.etree.ElementTree as Et
import logging
import hcpsdk


__all__ = ['ReplicationSettingsError', 'Replication']

logging.getLogger('hcpsdk.mapi.replication').addHandler(logging.NullHandler())


class ReplicationSettingsError(Exception):
    """
    Indicate an invalid action for the given link type (R_BEGINRECOVERY or
    R_COMPLETERECOVERY on a R_ACTIVE_ACTIVE link, R_FAILBACK on an
    R_OUTBOUND or R_INBOUND link).
    """

    def __init__(self, reason):
        """
        :param reason: An error description
        """
        self.args = (reason,)


class Replication(object):
    """
    Access replication link information, modify the replication link state.
    """
    # link types
    R_ACTIVE_ACTIVE = 'ACTIVE_ACTIVE'
    R_OUTBOUND = 'OUTBOUND'
    R_INBOUND = 'INBOUND'
    # link activity
    R_SUSPEND = 'suspend'  # suspend a link
    R_RESUME = 'resume'  # resume a link
    R_RESTORE = 'restore'  # restore a link
    R_FAILOVER = 'failOver'  # for all link types
    R_FAILBACK = 'failBack'  # for active/active links
    R_BEGINRECOVERY = 'beginRecovery'  # for active/passive links
    R_COMPLETERECOVERY = 'completeRecovery'  # dito

    def __init__(self, target, debuglevel=0):
        """
        :param target:      an hcpsdk.Target object
        :param debuglevel:  0..9 (used in *http.client*)
        """
        self.logger = logging.getLogger(__name__ + '.Replication')
        hcpsdk.checkport(target, hcpsdk.P_MAPI)
        self.target = target
        self.debuglevel = debuglevel
        self.connect_time = 0.0
        self.service_time = 0.0

    def getreplicationsettings(self):
        """
        Query MAPI for the general settings of the replication service.

        :return: a dict containing the settings
        :raises: HcpsdkError
        """
        d = {}
        try:
            con = hcpsdk.Connection(self.target, debuglevel=self.debuglevel)
        except Exception as e:
            raise hcpsdk.HcpsdkError(str(e))
        else:
            self.connect_time = con.connect_time
            try:
                r = con.GET('/mapi/services/replication')
            except Exception as e:
                raise hcpsdk.HcpsdkError(str(e))
            else:
                if r.status == 200:
                    # Good status, get and parse the Response
                    x = r.read()
                    self.service_time = con.service_time2
                    for child in Et.fromstring(x):
                        d[child.tag] = child.text
                else:
                    raise (hcpsdk.HcpsdkError('{} - {}'.format(r.status, r.reason)))
        finally:
            # noinspection PyUnboundLocalVariable
            con.close()

        return d

    def getlinklist(self):
        """
        Query MAPI for a list of replication links.

        :return:  a list with the names of replication links
        :raises: HcpsdkError
        """
        d = []
        try:
            con = hcpsdk.Connection(self.target, debuglevel=self.debuglevel)
        except Exception as e:
            raise hcpsdk.HcpsdkError(str(e))
        else:
            self.connect_time = con.connect_time
            try:
                r = con.GET('/mapi/services/replication/links')
            except Exception as e:
                d.append('Error: {}'.format(str(e)))
            else:
                if r.status == 200:
                    # Good status, get and parse the Response
                    x = r.read()
                    self.service_time = con.service_time2
                    root = Et.fromstring(x)
                    for child in root:
                        if child.tag == 'name':
                            d.append(child.text)
                else:
                    raise (hcpsdk.HcpsdkError('{} - {}'.format(r.status, r.reason)))
        finally:
            # noinspection PyUnboundLocalVariable
            con.close()

        return d

    def getlinkdetails(self, link):
        """
        Query MAPI for the details of a replication link.

        :param link:    the name of the link as retrieved by **getlinklist()**
        :return:        a dict holding the details
        :raises:        HcpsdkError
        """
        d = {}
        try:
            con = hcpsdk.Connection(self.target, debuglevel=self.debuglevel)
        except Exception as e:
            raise hcpsdk.HcpsdkError(str(e))
        else:
            self.connect_time = con.connect_time
            try:
                r = con.GET('/mapi/services/replication/links/{}'.format(link),
                            params={'verbose': 'true'})
            except Exception as e:
                raise hcpsdk.HcpsdkError(str(e))
            else:
                if r.status == 200:
                    # Good status, get and parse the Response
                    x = r.read()
                    self.service_time = con.service_time2
                    for child in Et.fromstring(x):
                        if child.text:
                            d[child.tag] = child.text
                        else:
                            d[child.tag] = {}
                            for i in child:
                                if i.text:
                                    d[child.tag][i.tag] = i.text
                                else:
                                    d[child.tag][i.tag] = {}
                                    for j in i:
                                        d[child.tag][i.tag][j.tag] = j.text
                else:
                    raise (hcpsdk.HcpsdkError('{} - {}'.format(r.status, r.reason)))
        finally:
            # noinspection PyUnboundLocalVariable
            con.close()

        return d

    def setreplicationlinkstate(self, linkname, action, linktype=None):
        """
        Alter the state of a replication link.

        :param linkname:    name of the link to change the state
        :param linktype:    one of ``[R_ACTIVE_ACTIVE, R_OUTBOUND, R_INBOUND]``;
                            not required for ``[R_SUSPEND, R_RESUME, R_RESTORE]``
        :param action:     one of ``[R_SUSPEND, R_RESUME, R_RESTORE, R_FAILOVER,
                            R_FAILBACK, R_BEGINRECOVERY, R_COMPLETERECOVERY]``
        :raises: HcpsdkError
        """
        # make sure that only valid linktypes and actions are accepted
        if linktype not in [Replication.R_ACTIVE_ACTIVE, Replication.R_OUTBOUND, Replication.R_INBOUND, None] or \
                        action not in [Replication.R_SUSPEND, Replication.R_RESUME,
                                       Replication.R_RESTORE, Replication.R_BEGINRECOVERY,
                                       Replication.R_COMPLETERECOVERY,
                                       Replication.R_FAILBACK, Replication.R_FAILOVER]:
            raise ValueError
        # make sure that no invalid action is called
        if (action == Replication.R_FAILBACK and linktype in [Replication.R_OUTBOUND, Replication.R_INBOUND]) or \
                (action in [Replication.R_BEGINRECOVERY, Replication.R_COMPLETERECOVERY] and
                         linktype == Replication.R_ACTIVE_ACTIVE) or \
                (action in [Replication.R_FAILOVER, Replication.R_FAILBACK,
                            Replication.R_BEGINRECOVERY, Replication.R_COMPLETERECOVERY] and not linktype):
            raise ReplicationSettingsError('{} not allowed on {} link'.format(action, linktype))

        # build params
        action = {action: ''}
        # let's do it!
        try:
            con = hcpsdk.Connection(self.target, debuglevel=self.debuglevel)
        except Exception as e:
            raise hcpsdk.HcpsdkError(str(e))
        else:
            try:
                r = con.POST('/mapi/services/replication/links/{}'
                             .format(linkname), params=action)
            except Exception as e:
                raise hcpsdk.HcpsdkError(str(e))
            else:
                if r.status != 200:
                    err = r.getheader('X-HCP-ErrorMessage', 'no message')
                    raise (hcpsdk.HcpsdkError('{} - {} ({})'.format(r.status, r.reason, err)))
                else:
                    r.read()
        finally:
            # noinspection PyUnboundLocalVariable
            con.close()





import sys
import hcpsdk

if __name__ == '__main__':

    #try:
    t = hcpsdk.target('n1.m.hcp73.archivas.com', 'n', 'n01',
                      replica_fqdn='n1.m.hcp72.archivas.com')
    #except hcpsdk.HcpsdkReplicaInitError as e:
    #    sys.exit(e)

    print('primary =',t.fqdn, t.addresses)
    print('replica =', t.replica.fqdn, t.replica.addresses)



# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Thorsten Simons (sw@snomis.de)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import unittest
from pprint import pprint

import hcpsdk
import init_tests as it


class LtVars(object):
    linkname = ''


class TestHcpsdk_40_1_Mapi_Replication(unittest.TestCase):
    def setUp(self):
        # self.T_NS_GOOD = "admin.hcp1.snomis.local"
        # self.T_USER = "service"
        # self.T_PASSWORD = "service01"
        # self.T_AUTH = hcpsdk.NativeAuthorization(self.T_USER, self.T_PASSWORD)
        # self.T_PORT = 9090
        self.hcptarget = hcpsdk.Target(it.P_ADMIN, it.P_ADMAUTH,
                                       port=it.P_MAPIPORT, dnscache=it.P_DNSCACHE)
        self.mapi = hcpsdk.mapi.Replication(self.hcptarget)

    def tearDown(self):
        del self.hcptarget

    def test_1_10_good_getReplicationSettings(self):
        """
        Make sure we get a list with at least one entry
        """
        print('test_1_10_good_getReplicationSettings:')
        r = self.mapi.getreplicationsettings()
        pprint(r)
        self.assertTrue(type(r) == dict)

    def test_1_20_good_getLinkList(self):
        """
        Make sure we get a list with at least one entry
        """
        print('test_1_20_good_getLinkList:')
        r = self.mapi.getlinklist()
        print(r)
        self.assertTrue(type(r) == list)
        self.assertTrue(len(r) >= 1)
        LtVars.linkname = r[0]

    def test_1_30_good_getLinkDetails(self):
        """
        Make sure we get a list with at least one entry
        """
        print('test_1_30_good_getLinkDetails({}):'.format(LtVars.linkname))
        r = self.mapi.getlinkdetails(link=LtVars.linkname)
        pprint(r)
        self.assertTrue(True)


if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Thorsten Simons (sw@snomis.de)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import threading
import socket
import logging
# noinspection PyPackageRequirements
import dns
# noinspection PyPackageRequirements
import dns.resolver


__all__ = ['IpsError', 'Circle', 'Request', 'Response', 'query']

logging.getLogger('hcpsdk.ips').addHandler(logging.NullHandler())


class IpsError(Exception):
    """
    Signal an error in *hcpsdk.ips* - typically a name resolution problem.
    """

    def __init__(self, reason):
        """
        :param reason:  an error description
        """
        self.args = (reason,)


# noinspection PyTypeChecker
class Circle(object):
    """
    Resolve an FQDN (using **query()**), cache the acquired IP addresses and
    yield them round-robin.
    """
    __EMPTY_ADDRLIST = []

    def __init__(self, fqdn, port=443, dnscache=False):
        """
        :param fqdn:        the FQDN to be resolved
        :param port:        the port to be used by the **hcpsdk.Target** object
        :param dnscache:    if True, use the system resolver (which **might** do
                            local caching), else use an internal resolver,
                            bypassing any cache available
        :returns:           an *hcpsdk.ips.Response* object
        """
        self.logger = logging.getLogger(__name__ + '.Circle')
        self.__authority = fqdn
        self.__port = port
        self.__dnscache = dnscache
        self._cLock = threading.Lock()
        self.__generator = None
        self._addresses = Circle.__EMPTY_ADDRLIST.copy()
        self.logger = logging.getLogger('hcpsdk.ips.Circle')

        # initial lookup, build the address cache
        self._addr(fqdn=self.__authority)

    def _addr(self, fqdn=None):
        """
        If called with a dnsname (FQDN), query DNS for that name,
        cache the acquired IP addresses.
        If called without dnsname, work as a generator that yields the
        cached IP addresses in a round-robin fashion.

        .. Warning::
            This method is intended to be internal to **hcpsdk** and may be used
            from the outside *without* parameters, only.

        :param fqdn:    the FQDN
        :return:        an IP address (as string)
        """

        def __addr(dnsname, dnscache=False):
            """
            resolve HCPs IP addresses and build a list with all IPs gathered
            """
            self._addresses = Circle.__EMPTY_ADDRLIST.copy()
            result = query(dnsname, cache=dnscache)
            if result.raised:
                raise IpsError(result.raised)
            self._addresses = result.ips.copy()

            while True:
                for ipadr in self._addresses:
                    yield str(ipadr)

        # acquire a lock to make sure that one Request gets serviced at a time
        self._cLock.acquire()
        if fqdn:
            self.__generator = __addr(fqdn, dnscache=self.__dnscache)
        myaddr = next(self.__generator)
        self._cLock.release()
        if fqdn:
            self.logger.debug('(re-) loaded IP address cache: {}, dnscache = {}'
                              .format(self._addresses, self.__dnscache))
        self.logger.debug('issued IP address: {}'.format(myaddr))
        return myaddr

    def refresh(self):
        """
        Force a fresh DNS query and rebuild the cached list of IP addresses
        """
        self._addr(fqdn=self.__authority)
        self.logger.debug('IP address cache refreshed')

    def __getattr__(self, item):
        """
        Used to make _addresses a read-only attributes
        """
        if item == '_addresses':
            return self._addresses
        else:
            raise AttributeError

    def __str__(self):
        return self.answer.qname


class Request(object):
    """
    A DNS query Request object
    """

    def __init__(self, fqdn, cache):
        self.fqdn = fqdn
        self.cache = cache
        self.sub = None


class Response(object):
    """
    DNS query Response object, returned by the **query()** function.
    """

    def __init__(self, fqdn, cache):
        """
        :param fqdn:    the FQDN for the Response
        :param cache:   Response from a query by-passing the local DNS cache (False)
                        or using the system resolver (True)
        """
        self.fqdn = fqdn
        self.cache = cache
        self.ips = []
        self.raised = ''


def query(fqdn, cache=False):
    """
    Submit a DNS query, using *socket.getaddrinfo()* if cache=True, or
    *dns.resolver.query()* if cache=False.

    :param fqdn:    a FQDN to query DNS -or- a *Request* object
    :param cache:   if True, use the system resolver (which might do local caching),
                    else use an internal resolver, bypassing any cache available
    :return:        an **hcpsdk.ips.Response** object
    :raises:        should never raise, as Exceptions are signaled through
                    the **Response.raised** attribute
    """
    if isinstance(fqdn, Request):
        _response = Response(fqdn.fqdn, fqdn.cache)  # to collect the resolved IP addresses
    else:
        _response = Response(fqdn, cache)  # to collect the resolved IP addresses

    if _response.cache:
        try:
            ips = socket.getaddrinfo(_response.fqdn, 443, family=socket.AF_INET, type=socket.SOCK_DGRAM)
        except Exception as e:
            _response.raised = 'Err: ' + str(e)
        else:
            for a in ips:
                if a[4][0] not in _response.ips:
                    _response.ips.append(a[4][0])
    else:
        try:
            ips = dns.resolver.query(_response.fqdn, raise_on_no_answer=True)
        except dns.resolver.NXDOMAIN:
            _response.raised = 'Err: NXDOMAIN - The query name does not exist.'
        except dns.resolver.YXDOMAIN:
            _response.raised = 'Err: The query name is too long after DNAME substitution.'
        except dns.resolver.Timeout:
            _response.raised = 'Err: The operation timed out.'
        except dns.resolver.NoAnswer:
            _response.raised = 'Err: The Response did not contain an answer to the question.'
        except dns.resolver.NoNameservers:
            _response.raised = 'Err: NoNameservers - No non-broken nameservers are available to answer the query.'
        except dns.resolver.NotAbsolute:
            _response.raised = 'Err: Raised if an absolute domain name is required but a relative name was provided.'
        except dns.resolver.NoRootSOA:
            _response.raised = 'Err: Raised if for some reason there is no SOA at the root name. ' \
                               'This should never happen!'
        except dns.resolver.NoMetaqueries:
            _response.raised = 'Err: Metaqueries are not allowed.'
        except Exception as e:
            _response.raised = 'Err: ' + str(e)
        else:
            for i in ips.rrset:
                ip = str(i)
                if ip[1] == '#':
                    hx = ip[-8:]
                    ip = '{}.{}.{}.{}'.format(int(hx[:2], 16), int(hx[2:4], 16),
                                              int(hx[4:6], 16), int(hx[6:], 16))
                _response.ips.append(str(ip))
            if not len(_response.ips):
                _response.raised = 'Err: no Response'

    return _response




# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Thorsten Simons (sw@snomis.de)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import unittest
from collections import OrderedDict
from pprint import pprint

import hcpsdk
import init_tests as it

import logging
logging.basicConfig(filename='_example.log', level=logging.DEBUG)


class TestHcpsdk_30_1_NamespaceInfo_NS_GOOD(unittest.TestCase):
    def setUp(self):
        # self.T_NS_GOOD = "n1.m.hcp1.snomis.local"
        # self.T_USER = "n"
        # self.T_PASSWORD = "n01"
        # self.T_AUTH = hcpsdk.NativeAuthorization(self.T_USER, self.T_PASSWORD)
        # self.T_PORT = 443
        self.hcptarget = hcpsdk.Target(it.P_NS_GOOD, it.P_AUTH, port=it.P_PORT, dnscache=it.P_DNSCACHE)
        self.nso = hcpsdk.namespace.Info(self.hcptarget)

    def tearDown(self):
        del self.hcptarget

    # @unittest.skip("demonstrating skipping")
    def test_1_10_good_NSstatistics(self):
        """
        Make sure we get a dict w/ length 9
        """
        print('test_1_10_good_NSstatistics')
        r = self.nso.nsstatistics()
        pprint(r)
        self.assertTrue(type(r) == dict)
        self.assertTrue(len(r) == 9)

    # @unittest.skip("demonstrating skipping")
    def test_1_20_good_listAccessibleNS(self):
        """
        Make sure we get a dict holding dicts
        """
        print('test_1_20_good_listAccessibleNS')
        r = self.nso.listaccessiblens()
        pprint(r)
        self.assertTrue(type(r) == OrderedDict)
        for ns in r:
            self.assertTrue(type(r[ns]) == dict)

    def test_1_30_good_listThisNSonly(self):
        print('test_1_30_good_listThisNSonly')
        r = self.nso.listaccessiblens(all=False)
        pprint(r)
        self.assertTrue(type(r) == OrderedDict)
        self.assertTrue(len(r) == 1)
        for ns in r:
            self.assertTrue(type(r[ns]) == dict)

        #    @unittest.skip("demonstrating skipping")

    def test_1_40_good_listRetentionClasses(self):
        r = self.nso.listretentionclasses()
        pprint(r)
        self.assertTrue(type(r) == OrderedDict)
        for ns in r:
            self.assertTrue(type(r[ns]) == dict)

        #    @unittest.skip("demonstrating skipping")

    def test_1_50_good_listPermissions(self):
        print('test_1_50_good_listPermissions')
        r = self.nso.listpermissions()
        pprint(r)
        self.assertTrue(type(r) == OrderedDict)
        for ns in r:
            self.assertTrue(type(r[ns]) == dict)
        for nsk in r.keys():
            self.assertTrue(nsk in ['namespacePermissions',
                                    'namespaceEffectivePermissions',
                                    'userPermissions',
                                    'userEffectivePermissions'])


if __name__ == '__main__':
    unittest.main()


import sys
from os.path import join, normpath
import concurrent.futures
import time
import pprint
import hcpsdk

T_NAMESPACE = 'n1.m.hcp73.archivas.com'
T_DAAC      = 'n'
T_DAAC_PWD  = 'n01'
T_PORT      = 443
T_DEBUG     = True

T_THREADS        = 1          # no. of threads to use in parallel
T_OBJ            = 128            # the object we will use in a run (from objs)
T_CNT_PER_THREAD = 10          # how often the object will be written
T_OUTNAME   = '_hcp_loadtest'   # filename to be used
T_STARTPATH = '/rest/_hcp_loadtest/'    # path to write to

# a names of the source objects
T_OBJS = {     1: '1kbfile',
               8: '8kbfile',
              16: '16kbfile',
              32: '32kbfile',
              64: '64kbfile',
             128: '128kbfile',
             256: '256kbfile',
             512: '512kbfile',
            1024: '1mbfile',
           10240: '10mbfile',
          102400: '100mbfile'}


def t_write(connection, tgtpath, buffer):
    '''
    Write a set of objects
    :param connection: the http connection to be used
    :param tgtpaths: a list of target paths
    :param buffer: the bytesarray holding the object
    :return: (overall service tme, no. of objs processed)
    '''
    try:
        r = connection.PUT(tgtpath, buffer)
    except Exception as e:
        print('Exception: {}'.format(str(e)))
    else:
        connection.read()

    return

def t_read(connection, tgtpath):
    '''
    Read a set of objects
    :param buffer: the bytesarray holding the object
    :param connection: the http connection to be used
    :param tgtpaths: a list of target paths
    :return: (overall service tme, no. of objs processed)
    '''
    try:
        r = connection.GET(tgtpath)
    except Exception as e:
        print('Exception: {}'.format(str(e)))
    else:
        connection.read()

    return

def t_delete(connection, tgtpath):
    '''
    Delete a set of objects
    :param buffer: the bytesarray holding the object
    :param connection: the http connection to be used
    :param tgtpaths: a list of target paths
    :return: (overall service tme, no. of objs processed)
    '''
    try:
        r = connection.DELETE(tgtpath)
    except Exception as e:
        print('Exception: {}'.format(str(e)))
    else:
        connection.read()

    return



if __name__ == '__main__':
    if T_DEBUG:
        import logging
        logging.basicConfig(level=logging.DEBUG, style='{', format='{levelname:>5s} {msg}')
        # noinspection PyShadowingBuiltins
        print = pprint = logging.info

    # the place where we will find the needed source objects
    if sys.platform == 'win32':
        T_INPATH = 'd:\\__files'
    elif sys.platform == 'darwin':
        T_INPATH = '/Volumes/dev/__files'
    else:
        sys.exit('source path is undefined')

    print("--> Init <hcptarget> object")
    try:
        auth = hcpsdk.NativeAuthorization(T_DAAC, T_DAAC_PWD)
        hcptarget = hcpsdk.Target(T_NAMESPACE, auth, T_PORT)
    except hcpsdk.HcpsdkError as e:
        sys.exit("Fatal: {}".format(e.errText))

    print("--> Init {} <connection> object(s)".format(T_THREADS))
    conns = None
    conntimes = 0.0
    outer_contime = time.time()
    try:
        conns = hcpsdk.Connection(hcptarget, debuglevel=0, idletime=3)
    except Exception as e:
        sys.exit('Exception: {}'.format(str(e)))
    else:
        conntimes += conns.connect_time
    outer_contime = time.time() - outer_contime

    print('\tinner connection time: {:,.5f}'.format(conntimes))
    print('\touter connection time: {:,.5f}'.format(outer_contime))

    # load one of the files into memory
    print('--> Loading test object into memory...')
    with open(join(T_INPATH,T_OBJS[T_OBJ]), 'rb') as inHdl:
        buf = bytearray(inHdl.read())
        size = len(buf)
        print('...object size = {}'.format(size))

    # Create the output paths
    outpath = join(T_STARTPATH, 'xxxxx')

    # now write the objects...
    #-------------------------
    print('--> Now ingesting objects:')
    outer_contime = time.time()
    try:
        t_write(conns, outpath, buf)
        print('PUT status: {}'.format(conns._response.status))
    except Exception as exc:
        print('PUT generated an exception: {}'.format(exc))
    outer_contime = time.time() - outer_contime
    print('\tinner service time:  {:,.5f} ({:5,.2f} Mb/sec)'.format(conns.service_time2,
                                                                    size/conns.service_time2/1024/1024))
    print('\touter service time:  {:,.5f} ({:5,.2f} Mb/sec)'.format(outer_contime, size/outer_contime/1024/1024))

    print('sleeping 10 secs')
    time.sleep(10)

    # now reading the objects...
    #---------------------------
    print('--> Now reading objects:')
    outer_contime = time.time()
    try:
        t_read(conns, outpath)
        print('GET status: {}'.format(conns._response.status))
    except Exception as exc:
        print('GET generated an exception: {}'.format('...', exc))
    outer_contime = time.time() - outer_contime
    print('\tinner service time:  {:,.5f} ({:5,.2f} Mb/sec)'.format(conns.service_time2,
                                                                    size/conns.service_time2/1024/1024))
    print('\touter service time:  {:,.5f} ({:5,.2f} Mb/sec)'.format(outer_contime, size/outer_contime/1024/1024))

    # now deleting objects...
    print('--> Now deleting objects:')
    outer_contime = time.time()
    try:
        t_delete(conns, outpath)
        print('DELETE status: {}'.format(conns._response.status))
    except Exception as exc:
        print('DELETE generated an exception: {}'.format('...', exc))
    outer_contime = time.time() - outer_contime
    print('\tinner service time:  {:,.5f} ({:5,.2f} Mb/sec)'.format(conns.service_time2,
                                                                    size/conns.service_time2/1024/1024))
    print('\touter service time:  {:,.5f} ({:5,.2f} Mb/sec)'.format(outer_contime, size/outer_contime/1024/1024))


    # close the session
    conns.close()



# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Thorsten Simons (sw@snomis.de)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
import os.path
sys.path.insert(0, os.path.abspath('..'))
import hcpsdk
print(hcpsdk.version())
import unittest
import ssl
from pprint import pprint

import init_tests as it


# @unittest.skip("skip TestHcpsdk_3_Access_https_certfile")
class TestHcpsdk_1_Access_https_certfile(unittest.TestCase):
    '''
    Make sure we can write/head/post/delete a file using https,
    verifying the cert against a local rootCertificate.
    '''
    def setUp(self):
        print('>>> TestHcpsdk_1_Access_https_certfile:')
        self.T_HCPFILE = '/rest/hcpsdk/TestHCPsdk_20_access'
        self.ctxt = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH,
                                               cafile='certs/rootCertificate.pem')
        print('Certificate store status:')
        pprint(self.ctxt.cert_store_stats())
        print('CA certificates:')
        pprint((self.ctxt.get_ca_certs()))
        self.hcptarget = hcpsdk.Target(it.P_NS_GOOD, it.P_AUTH, it.P_SSLPORT,
                                       dnscache=it.P_DNSCACHE, sslcontext=self.ctxt)
        self.con = hcpsdk.Connection(self.hcptarget)

    def tearDown(self):
        self.con.close()
        del self.hcptarget

    def test_1_10_put(self):
        """
        Ingest a file
        """
        # noinspection PyPep8Naming
        T_BUF = '0123456789ABCDEF' * 64
        r = self.con.PUT(self.T_HCPFILE, T_BUF)
        self.assertEqual(r.status, 201)

    def test_1_20_head(self):
        """
        Delete a file
        """
        r = self.con.HEAD(self.T_HCPFILE)
        self.assertEqual(r.status, 200)

    def test_1_30_post(self):
        """
        Delete a file
        """
        r = self.con.POST(self.T_HCPFILE, {'index': 'true'})
        self.assertEqual(r.status, 200)

    def test_1_90_delete(self):
        """
        Delete a file
        """
        r = self.con.DELETE(self.T_HCPFILE)
        self.assertEqual(r.status, 200)


# @unittest.skip("skip TestHcpsdk_4_Access_https_systemCA")
class TestHcpsdk_2_Access_https_systemCA(unittest.TestCase):
    '''
    Make sure we can write/head/post/delete a file using https,
    verifying the cert against the systems CA store.
    '''
    def setUp(self):
        print('>>> TestHcpsdk_2_Access_https_systemCA:')
        self.T_HCPFILE = '/rest/hcpsdk/TestHCPsdk_20_access'
        self.ctxt = ssl.create_default_context()
        print('Certificate store status:')
        pprint(self.ctxt.cert_store_stats())
        print('CA certificates:')
        pprint((self.ctxt.get_ca_certs()))
        self.hcptarget = hcpsdk.Target(it.P_NS_GOOD, it.P_AUTH, it.P_SSLPORT,
                                       dnscache=it.P_DNSCACHE, sslcontext=self.ctxt)
        self.con = hcpsdk.Connection(self.hcptarget)

    def tearDown(self):
        self.con.close()
        del self.hcptarget

    def test_2_10_put(self):
        """
        Ingest a file
        """
        # noinspection PyPep8Naming
        T_BUF = '0123456789ABCDEF' * 64
        r = self.con.PUT(self.T_HCPFILE, T_BUF)
        print(self.con.response_status, self.con.response_reason)
        self.assertEqual(r.status, 201)

    def test_2_90_delete(self):
        """
        Delete a file
        """
        r = self.con.DELETE(self.T_HCPFILE)
        print(self.con.response_status, self.con.response_reason)
        self.assertEqual(r.status, 200)


# @unittest.skip("demonstrating skipping")
class TestHcpsdk_3_Access_https_certfile_fail(unittest.TestCase):
    '''
    Make sure we fail write/delete a file using https,
    verifying the cert against a false rootCertificate.
    '''
    def setUp(self):
        print('>>> TestHcpsdk_3_Access_https_certfile_fail:')
        self.T_HCPFILE = '/rest/hcpsdk/TestHCPsdk_20_access'
        self.ctxt = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH,
                                               cafile='certs/failCertificate.pem')
        print('Certificate store status:')
        pprint(self.ctxt.cert_store_stats())
        print('CA certificates:')
        pprint((self.ctxt.get_ca_certs()))
        self.hcptarget = hcpsdk.Target(it.P_NS_GOOD, it.P_AUTH, it.P_SSLPORT,
                                       dnscache=it.P_DNSCACHE, sslcontext=self.ctxt)
        self.con = hcpsdk.Connection(self.hcptarget)

    def tearDown(self):
        self.con.close()
        del self.hcptarget

    def test_3_10_put(self):
        """
        Ingest a file
        """
        # noinspection PyPep8Naming
        T_BUF = '0123456789ABCDEF' * 64
        with self.assertRaises(hcpsdk.HcpsdkCertificateError):
            r = self.con.PUT(self.T_HCPFILE, T_BUF)

    def test_3_90_delete(self):
        """
        Delete a file
        """
        with self.assertRaises(hcpsdk.HcpsdkCertificateError):
            r = self.con.DELETE(self.T_HCPFILE)


# @unittest.skip("demonstrating skipping")
class TestHcpsdk_4_Access_Fail(unittest.TestCase):
    '''
    Make sure we fail when trying to read a non-existant file
    not verifying the cert at all.
    '''
    def setUp(self):
        self.T_HCPFILE = '/rest/hcpsdk/fail_TestHCPsdk_20_access'
        self.hcptarget = hcpsdk.Target(it.P_NS_GOOD, it.P_AUTH, it.P_SSLPORT, dnscache=it.P_DNSCACHE)
        self.con = hcpsdk.Connection(self.hcptarget)

    def tearDown(self):
        self.con.close()
        del self.hcptarget

    def test_4_10_get(self):
        """
        Ingest a file
        """
        self.r = self.con.GET(self.T_HCPFILE)
        print(self.con.response_status, self.con.response_reason)
        self.assertEqual(self.r.status, 404)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Thorsten Simons (sw@snomis.de)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import unittest
import logging
import sys
from pprint import pprint

import hcpsdk
import init_tests as it


class TestHcpsdk_43_1_Mapi_Tenant(unittest.TestCase):
    def setUp(self):
        sh = logging.StreamHandler(sys.stderr)
        fh = logging.Formatter("[%(levelname)-8s]: %(message)s")
        sh.setFormatter(fh)
        log = logging.getLogger()
        log.addHandler(sh)
        log.setLevel(logging.INFO)

        self.hcptarget = hcpsdk.Target(it.P_ADMIN, it.P_ADMAUTH,
                                       port=it.P_MAPIPORT, dnscache=False)

    def tearDown(self):
        del self.hcptarget

    def test_1_10_tenant_list(self):
        """
        Check if we can list all Tenants
        """
        print('test_1_10_tenant_list:')

        tenants = hcpsdk.mapi.listtenants(self.hcptarget, debuglevel=0)
        self.assertTrue(type(tenants) == list)

        for i in tenants:
            self.assertTrue(type(i) == hcpsdk.mapi.Tenant)
            pprint(i.info())
            i.close()




# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Thorsten Simons (sw@snomis.de)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import logging
from json import loads
from pprint import pprint
import hcpsdk


__all__ = ['TenantError', 'listtenants', 'Tenant']

logging.getLogger('hcpsdk.mapi.tenant').addHandler(logging.NullHandler())


class TenantError(Exception):
    """
    Base Exception used by the *hcpsdk.mapi.Tenant()* class
    """
    def __init__(self, reason):
        """
        :param reason: An error description
        """
        self.args = (reason,)


def listtenants(target, timeout=60, debuglevel=0):
    """
    Get a list of available Tenants

    :param target:      an hcpsdk.Target object
    :param timeout:     the connection timeout in seconds
    :param debuglevel:  0..9 (used in *http.client*)
    :returns:           a list() of *Tenant()* objects
    :raises:            *hcpsdk.HcpsdkPortError* in case *target* is
                        initialized with a port different that *P_MAPI*
    """
    logger = logging.getLogger(__name__)
    logger.debug('getting a list of Tenants')
    hcpsdk.checkport(target, hcpsdk.P_MAPI)
    tenantslist = []

    try:
        con = hcpsdk.Connection(target, timeout=timeout, debuglevel=debuglevel)
    except Exception as e:
        raise hcpsdk.HcpsdkError(str(e))

    try:
        con.GET('/mapi/tenants', headers={'Accept': 'application/json'},
                params={'verbose': 'true'})
    except Exception as e:
        logger.debug('getting a list of Tenants failed: {}'.format(e))
        raise TenantError('get Tenant list failed: {}'.format(e))
    else:
        if con.response_status == 200:
            for t in loads(con.read().decode())['name']:
                tenantslist.append(Tenant(target, t, debuglevel=debuglevel))
            logger.debug('got a list of {} Tenants'.format(len(tenantslist)))
        else:
            con.close()
            logger.debug('getting a list of Tenants failed: {}-{}'
                         .format(con.response_status, con.response_reason))
            raise TenantError('unable to list Tenants ({} - {})'
                              .format(con.response_status,
                                      con.response_reason))

    con.close()
    return tenantslist



class Tenant(object):
    """
    A class representing a Tenant
    """

    # TODO: this object is simply a container for a single tenants settings
    #       remove all the other stuff...

    def __init__(self, target, name, timeout=60, debuglevel=0):
        """
        :param target:      an hcpsdk.Target object
        :param name:        the Tenants name
        :param timeout:     the connection timeout in seconds
        :param debuglevel:  0..9 (used in *http.client*)
        """
        self.logger = logging.getLogger(__name__ + '.Tenant')
        self.target = target
        try:
            self.con = hcpsdk.Connection(self.target, timeout=timeout,
                                         debuglevel=debuglevel)
        except Exception as e:
            raise hcpsdk.HcpsdkError(str(e))

        self.name = name        # the Tenants name
        self._settings = {}     # the Tenants base settings

        self.logger.debug('initialized for "{}"'.format(self.name))


    def info(self, cache=True):
        """
        Get the settings of the Tenant

        :param cache:   a bool indicating if cached information shall be used
        :return:        a dict holding the Tenants settings
        """

        if not self._settings or not cache:
            try:
                self.con.GET('/mapi/tenants/{}'.format(self.name),
                             headers={'Accept': 'application/json'})
            except Exception as e:
                raise hcpsdk.HcpsdkError(str(e))
            else:
                if self.con.response_status == 200:
                    self._settings = loads(self.con.read().decode())
                    self.logger.debug('got settings of Tenant {}'
                                      .format(self.name))
                else:
                    self.logger.debug('getting settings of Tenant {} '
                                      'failed: {}-{}'
                                      .format(self.con.response_status,
                                              self.con.response_reason))
                    raise TenantError('unable to list Tenants ({} - {})'
                                      .format(self.con.response_status,
                                              self.con.response_reason))
        return self._settings

    def close(self):
        """
        Close the underlying *hcpsdk.Connection()* object
        """
        self.con.close()



import sys
from os.path import join, normpath
import concurrent.futures
import time
import pprint
import hcpsdk

T_NAMESPACE = 'n1.m.hcp73.archivas.com'
T_DAAC      = 'n'
T_DAAC_PWD  = 'n01'
T_PORT      = 443

T_THREADS        = 10          # no. of threads to use in parallel
T_OBJ            = 128            # the object we will use in a run (from objs)
T_CNT_PER_THREAD = 100          # how often the object will be written
T_OUTNAME   = '_hcp_loadtest'   # filename to be used
T_STARTPATH = '/rest/_hcp_loadtest/'    # path to write to

# a names of the source objects
T_OBJS = {     1: '1kbfile',
               8: '8kbfile',
              16: '16kbfile',
              32: '32kbfile',
              64: '64kbfile',
             128: '128kbfile',
             256: '256kbfile',
             512: '512kbfile',
            1024: '1mbfile',
           10240: '10mbfile',
          102400: '100mbfile'}


def t_write(connection, tgtpaths, buffer):
    '''
    Write a set of objects
    :param connection: the http connection to be used
    :param tgtpaths: a list of target paths
    :param buffer: the bytesarray holding the object
    :return: (overall service tme, no. of objs processed)
    '''
    sumtime = 0.0
    sumobjs = 0
    for i in tgtpaths:
        try:
            r = connection.request('PUT', i, buffer)
        except Exception as e:
            print('Exception: {}'.format(str(e)))
        else:
            connection.read()
            sumtime += connection.service_time2
            sumobjs += 1

    return(sumtime, sumobjs)

def t_read(connection, tgtpaths):
    '''
    Read a set of objects
    :param buffer: the bytesarray holding the object
    :param connection: the http connection to be used
    :param tgtpaths: a list of target paths
    :return: (overall service tme, no. of objs processed)
    '''
    sumtime = 0.0
    sumobjs = 0
    for i in tgtpaths:
        try:
            r = connection.request('GET', i)
        except Exception as e:
            print('Exception: {}'.format(str(e)))
        else:
            connection.read()
            sumtime += connection.service_time2
            sumobjs += 1

    return(sumtime, sumobjs)

def t_delete(connection, tgtpaths):
    '''
    Delete a set of objects
    :param buffer: the bytesarray holding the object
    :param connection: the http connection to be used
    :param tgtpaths: a list of target paths
    :return: (overall service tme, no. of objs processed)
    '''
    sumtime = 0.0
    sumobjs = 0
    for i in tgtpaths:
        try:
            r = connection.request('DELETE', i)
        except Exception as e:
            print('Exception: {}'.format(str(e)))
        else:
            connection.read()
            sumtime += connection.service_time2
            sumobjs += 1

    return(sumtime, sumobjs)



if __name__ == '__main__':
    # the place where we will find the needed source objects
    if sys.platform == 'win32':
        T_INPATH = 'd:\\__files'
    elif sys.platform == 'darwin':
        T_INPATH = '/Volumes/dev/__files'
    else:
        sys.exit('source path is undefined')

    # the paths in which we will store the objects
    T_PATHS = {}
    for i in range (0, 256):
        T_PATHS[i] = normpath('{}/p{:03}'.format(T_STARTPATH, i))

    print("--> Init <hcptarget> object")
    try:
        hcptarget = hcpsdk.target(T_NAMESPACE, T_DAAC, T_DAAC_PWD, T_PORT)
    except hcpsdk.HcpsdkError as e:
        sys.exit("Fatal: {}".format(e.errText))

    input('==> Press Enter to start the process: ')

    print("--> Init {} <connection> object(s)".format(T_THREADS))
    conns = []
    conntimes = 0.0
    outer_contime = time.time()
    for i in range(0, T_THREADS):
        try:
            conns.append(hcpsdk.connection(hcptarget, debuglevel=0))
        except Exception as e:
            sys.exit('Exception: {}'.format(str(e)))
        else:
            conntimes += conns[i].connect_time
    outer_contime = time.time() - outer_contime

    print('\tAverage inner connection time: {:,.5f}'.format(conntimes/T_THREADS))
    print('\tAverage outer connection time: {:,.5f}'.format(outer_contime))

    # load one of the files into memory
    print('--> Loading test object into memory - ', end='')
    with open(join(T_INPATH,T_OBJS[T_OBJ]), 'rb') as inHdl:
        buf = bytearray(inHdl.read())
        size = len(buf)
        print('object size = {}'.format(size))

    # Create the output paths
    outpaths = {}
    for i in range(0, T_THREADS):
        outpaths[i] = []
        for j in range(0, T_CNT_PER_THREAD):
            outpaths[i].append(join(T_STARTPATH, 't-{:03}'.format(i), '{}.{:03}'.format(T_OUTNAME, j)))

    # now write the objects...
    print('--> Now ingesting objects:')
    sumtime = 0.0
    sumobjs = 0
    outer_contime = time.time()
    with concurrent.futures.ThreadPoolExecutor(max_workers=T_THREADS) as executor:
        f = {executor.submit(t_write, conns[i], outpaths[i], buf): i for i in range(0,T_THREADS)}
        for future in concurrent.futures.as_completed(f):
            try:
                (stime, sobjs) = future.result()
            except Exception as exc:
                print('t-{} generated an exception: {}'.format('...', exc))
            else:
                sumtime += stime
                sumobjs += sobjs
    outer_contime = time.time() - outer_contime
    print('\t        Inner summary time:  {:,.5f} ({} objects)'.format(sumtime, sumobjs))
    print('\tAverage inner service time:  {:,.5f} ({:5,.2f} Mb/sec)'.format(sumtime / T_THREADS / sumobjs,
                                                                          (T_THREADS*sumobjs*size)/sumtime/1024/1024),
          flush=True)
    print('\tAverage outer service time:  {:,.5f} ({:5,.2f} Mb/sec)'.format(outer_contime,
                                                                          (T_THREADS*sumobjs*size)/outer_contime/1024/1024),
          flush=True)

    # now reading the objects...
    print('--> Now reading objects:')
    sumtime = 0.0
    sumobjs = 0
    outer_contime = time.time()
    with concurrent.futures.ThreadPoolExecutor(max_workers=T_THREADS) as executor:
        f = {executor.submit(t_read, conns[i], outpaths[i]): i for i in range(0,T_THREADS)}
        for future in concurrent.futures.as_completed(f):
            try:
                (stime, sobjs) = future.result()
            except Exception as exc:
                print('t-{} generated an exception: {}'.format('...', exc))
            else:
                sumtime += stime
                sumobjs += sobjs
    outer_contime = time.time() - outer_contime
    print('\t        Inner summary time:  {:,.5f} ({} objects)'.format(sumtime, sumobjs))
    print('\tAverage inner service time:  {:,.5f} ({:5,.2f} Mb/sec)'.format(sumtime / T_THREADS / sumobjs,
                                                                          (T_THREADS*sumobjs*size)/sumtime/1024/1024),
          flush=True)
    print('\tAverage outer service time:  {:,.5f} ({:5,.2f} Mb/sec)'.format(outer_contime,
                                                                          (T_THREADS*sumobjs*size)/outer_contime/1024/1024),
          flush=True)

    # now deleting objects...
    print('--> Now deleting objects:')
    sumtime = 0.0
    sumobjs = 0
    outer_contime = time.time()
    with concurrent.futures.ThreadPoolExecutor(max_workers=T_THREADS) as executor:
        f = {executor.submit(t_delete, conns[i], outpaths[i]): i for i in range(0,T_THREADS)}
        for future in concurrent.futures.as_completed(f):
            try:
                (stime, sobjs) = future.result()
            except Exception as exc:
                print('t-{} generated an exception: {}'.format('...', exc))
            else:
                sumtime += stime
                sumobjs += sobjs
    outer_contime = time.time() - outer_contime
    print('\t        Inner summary time:  {:,.5f} ({} objects)'.format(sumtime, sumobjs))
    print('\tAverage inner service time:  {:,.5f}'.format(sumtime / T_THREADS / sumobjs))
    print('\tAverage outer service time:  {:,.5f}'.format(outer_contime))


    # close the session
    for con in conns:
        con.close()



# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Thorsten Simons (sw@snomis.de)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
import os.path
sys.path.insert(0, os.path.abspath('..'))
import hcpsdk
from hcpsdk.ips import IpsError
import unittest
import ssl
import socket
import http.client
from pprint import pprint

import init_tests as it

print('hcpsdk: ', hcpsdk.version())



# @unittest.skip("skip TestHcpsdk_01_Target")
class TestHcpsdk_01_Target(unittest.TestCase):

    def test_01_10_ip_address_available(self):
        """
        Make sure we can get a single IP address from Target object's pool
        """
        hcptarget = hcpsdk.Target(it.P_NS_GOOD, it.P_AUTH, port=it.P_PORT,
                                  dnscache=it.P_DNSCACHE)
        self.assertTrue(hcptarget.getaddr() in hcptarget.addresses)

    def test_01_20_ip_address_not_available(self):
        """
        Make sure an exception is raised if the FQDN can't be resolved
        """
        with self.assertRaises(hcpsdk.ips.IpsError):
            hcpsdk.Target(it.P_NS_BAD, it.P_AUTH, port=it.P_PORT,
                          dnscache=it.P_DNSCACHE)

    def test_01_30_good_target_authority(self):
        """
        Make sure we get a hcpsdk.Target object
        """
        hcptarget = hcpsdk.Target(it.P_NS_GOOD, it.P_AUTH, port=it.P_PORT,
                                  dnscache=it.P_DNSCACHE)
        self.assertIs(type(hcptarget), hcpsdk.Target)


# @unittest.skip("skip TestHcpsdk_02_Access_http")
class TestHcpsdk_02_Access_http(unittest.TestCase):
    '''
    Make sure we can write/head/post/delete a file using http
    '''
    def setUp(self):
        self.T_HCPFILE = '/rest/hcpsdk/TestHCPsdk_20_access'
        self.hcptarget = hcpsdk.Target(it.P_NS_GOOD, it.P_AUTH, it.P_PORT,
                                       dnscache=it.P_DNSCACHE)
        self.con = hcpsdk.Connection(self.hcptarget)

    def tearDown(self):
        self.con.close()
        del self.hcptarget

    def test_02_10_put(self):
        """
        Ingest a file
        """
        # noinspection PyPep8Naming
        T_BUF = '0123456789ABCDEF' * 64
        r = self.con.PUT(self.T_HCPFILE, T_BUF)
        self.assertEqual(r.status, 201)

    def test_02_20_head(self):
        """
        Delete a file
        """
        r = self.con.HEAD(self.T_HCPFILE)
        self.assertEqual(r.status, 200)

    def test_02_30_post(self):
        """
        Delete a file
        """
        r = self.con.POST(self.T_HCPFILE, params={'index': 'true'})
        self.assertEqual(r.status, 200)

    def test_02_90_delete(self):
        """
        Delete a file
        """
        r = self.con.DELETE(self.T_HCPFILE)
        self.assertEqual(r.status, 200)


# @unittest.skip("skip TestHcpsdk_03_Access_Fail")
class TestHcpsdk_03_Access_Fail(unittest.TestCase):
    '''
    Make sure we fail when trying to read a non-existant file
    not verifying the cert at all.
    '''
    def setUp(self):
        self.T_HCPFILE = '/rest/hcpsdk/fail_TestHCPsdk_20_access'
        self.hcptarget = hcpsdk.Target(it.P_NS_GOOD, it.P_AUTH, it.P_SSLPORT,
                                       dnscache=it.P_DNSCACHE)
        self.con = hcpsdk.Connection(self.hcptarget)

    def tearDown(self):
        self.con.close()
        del self.hcptarget

    def test_03_10_get(self):
        """
        Read a file
        """
        self.r = self.con.GET(self.T_HCPFILE)
        self.assertEqual(self.r.status, 404)


# @unittest.skip("skip TestHcpsdk_10_Errors")
class TestHcpsdk_10_Errors(unittest.TestCase):
    '''
    Make sure we get the proper error if an ConnectionAbortedError is raised
    '''
    def setUp(self):
        self.T_HCPFILE = '/rest/hcpsdk/TestHCPsdk_20_access'
        self.hcptarget = hcpsdk.Target(it.P_NS_GOOD, it.P_AUTH, it.P_SSLPORT,
                                       dnscache=it.P_DNSCACHE)
        self.con = hcpsdk.Connection(self.hcptarget, retries=1)

    def tearDown(self):
        r = self.con.DELETE(self.T_HCPFILE)
        self.con.close()
        del self.hcptarget

    def test_10_09_IpsError(self):
        """
        Ingest a file while forcing an error
        """
        T_BUF = '0123456789ABCDEF' * 64
        r = self.con.PUT(self.T_HCPFILE, T_BUF)
        self.assertEqual(r.status, 201)

        self.con._fail = IpsError
        self.assertRaises(IpsError, self.con.HEAD, self.T_HCPFILE)

    def test_10_10_ConnectionAbortedError(self):
        """
        Ingest a file while forcing an error
        """
        T_BUF = '0123456789ABCDEF' * 64
        r = self.con.PUT(self.T_HCPFILE, T_BUF)
        self.assertEqual(r.status, 201)

        self.con._fail = ConnectionAbortedError
        r = self.con.HEAD(self.T_HCPFILE)
        self.assertEqual(r.status, 200)

    def test_10_11_TimeoutError(self):
        """
        Ingest a file while forcing an error
        """
        T_BUF = '0123456789ABCDEF' * 64
        r = self.con.PUT(self.T_HCPFILE, T_BUF)
        self.assertEqual(r.status, 201)

        self.con._fail = TimeoutError
        r = self.con.HEAD(self.T_HCPFILE)
        self.assertEqual(r.status, 200)

    def test_10_12_sockettimeout(self):
        """
        Ingest a file while forcing an error
        """
        T_BUF = '0123456789ABCDEF' * 64
        r = self.con.PUT(self.T_HCPFILE, T_BUF)
        self.assertEqual(r.status, 201)

        self.con._fail = socket.timeout
        r = self.con.HEAD(self.T_HCPFILE)
        self.assertEqual(r.status, 200)

    def test_10_13_httpclientResponseNotReady(self):
        """
        Ingest a file while forcing an error
        """
        T_BUF = '0123456789ABCDEF' * 64
        r = self.con.PUT(self.T_HCPFILE, T_BUF)
        self.assertEqual(r.status, 201)

        self.con._fail = http.client.ResponseNotReady
        r = self.con.HEAD(self.T_HCPFILE)
        self.assertEqual(r.status, 200)

    def test_10_14_HcpsdkCertificateError(self):
        """
        Ingest a file while forcing an error
        """
        T_BUF = '0123456789ABCDEF' * 64
        r = self.con.PUT(self.T_HCPFILE, T_BUF)
        self.assertEqual(r.status, 201)

        self.con._fail = ssl.SSLError
        self.assertRaises(hcpsdk.HcpsdkCertificateError, self.con.HEAD, self.T_HCPFILE)

    def test_10_15_httpclientHTTPException(self):
        """
        Ingest a file while forcing an error
        """
        T_BUF = '0123456789ABCDEF' * 64
        r = self.con.PUT(self.T_HCPFILE, T_BUF)
        self.assertEqual(r.status, 201)

        self.con._fail = http.client.HTTPException
        self.assertRaises(hcpsdk.HcpsdkError, self.con.HEAD, self.T_HCPFILE)



if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Thorsten Simons (sw@snomis.de)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from posixpath import join
from io import BytesIO
from uuid import uuid1
from xml.etree.ElementTree import ElementTree
from xml.etree.ElementTree import Element
import logging

__all__ = ['PathBuilderError', 'PathBuilder']

logging.getLogger('hcpsdk.pathbuilder').addHandler(logging.NullHandler())


class PathBuilderError(Exception):
    """
    Used to signal an error during unique object name generation or object
    name to path mapping.
    """
    def __init__(self, reason):
        """
        :param reason:  an error description
        """
        self.args = (reason,)


class PathBuilder(object):
    """
    Conversion of a filename into a unique object name and a proper path for HCPs
    needs. Re-conversion of a unique object name to the path where the object can
    be found in HCP.
    """

    def __init__(self, initialpath='/rest/hcpsdk', annotation=False):
        """
        :param initialpath: the leading part of the path
        :param annotation:  if True, create an XML structure to be used as custom
                            metadata annotation containing a tag with the
                            original filename of the object.
        """
        self.logger = logging.getLogger(__name__ + '.PathBuilder')
        self.leadingpath = initialpath
        self.annotation = annotation

    def getunique(self, filename):
        """
        Build a unique path / object name scheme.

        The path is build from **initialpath** given during class instantiation
        plus byte 4 and 3 of the UUID in hexadecimal.

        If **annotation** is True during class instantiation, there will be
        a third element in the returned tuple, containing an XML structure that
        can be used as custom metadata annotation for the object.

        :param filename:    the filename to be transformed
        :return:            a tuple consisting of path, unique object name and
                            -eventually- an annotation string.
        :raises:            hcpsdk.pathbuilder.pathbuilderError

        **Example:**

        ::

            >>> from hcpsdk.pathbuilder import PathBuilder
            >>> p = PathBuilder(initialpath='/rest/mypath', annotation=True)
            >>> o = p.getunique('testfile.txt')
            >>> o
            ('/rest/mypath/b4/ec', '8ac8ecb4-9f1e-11e4-a524-98fe94437d8c',
             '<?xml version=\'1.0\' encoding=\'utf-8\'?>
                <hcpsdk_fileobject
                    filename="testfile.txt"
                    path="/rest/mypath/b4/ec"
                    uuid="8ac8ecb4-9f1e-11e4-a524-98fe94437d8c"
             />')
            >>>
        """
        try:
            uuid = str(uuid1())
            path = join(self.leadingpath, uuid[6:8], uuid[4:6])
            if self.annotation:
                xml = ElementTree()
                e = Element('hcpsdk_fileobject', {'filename': filename,
                                                  'path': path,
                                                  'uuid': uuid})
                # noinspection PyProtectedMember
                xml._setroot(e)
                with BytesIO() as xmlstring:
                    xml.write(xmlstring, encoding="utf-8", method="xml", xml_declaration=True)
                    annotation = xmlstring.getvalue().decode()
        except Exception as e:
            raise PathBuilderError(str(e))

        if self.annotation:
            # noinspection PyUnboundLocalVariable
            return path, uuid, annotation
        else:
            return path, uuid

    def getpath(self, objectname):
        """
        From a unique object name, retrieve the path in which the object was stored.

        :param objectname:  an unique object name
        :return:            the full path to the object (including its name)
        :raises:            hcpsdk.pathbuilder.pathbuilderError

        **Example:**

        ::

            >>> p.getpath('8ac8ecb4-9f1e-11e4-a524-98fe94437d8c')
            '/rest/mypath/b4/ec/8ac8ecb4-9f1e-11e4-a524-98fe94437d8c'
            >>>
        """
        try:
            path = join(self.leadingpath, objectname[6:8], objectname[4:6], objectname)
        except Exception as e:
            raise PathBuilderError(str(e))

        return path



# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Thorsten Simons (sw@snomis.de)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
from base64 import b64encode
from hashlib import md5
# As of Python 3.4.3, http.client.HTTPSconnection() will default to verify
# presented certificates against the system's trusted CA chain. To enable
# the previous behaviour, we switch it off.
import ssl
try:
    SSL_NOVERIFY = ssl._create_unverified_context()
except (AttributeError, NameError):
    SSL_NOVERIFY = None
import socket
import http.client
from urllib.parse import urlencode, quote
import logging
import time
from threading import Timer

# noinspection PyProtectedMember
from .version import _Version

# If we install using pip, we run into an error if we don't have
# dnspython installed; that's why we accept the ImportError here.
try:
    from . import ips
except ImportError as e:
    print('ImportError: {} - install dnspython >= 1.15.0'.format(e),
          file=sys.stderr)

from . import httpclient
from . import namespace
from . import mapi
from . import pathbuilder


__all__ = ['Target', 'Connection', 'BaseAuthorization', 'DummyAuthorization',
           'NativeAuthorization', 'NativeADAuthorization',
           'LocalSwiftAuthorization', 'HcpsdkError',
           'HcpsdkCantConnectError', 'HcpsdkTimeoutError',
           'HcpsdkCertificateError', 'HcpsdkReplicaInitError']

logging.getLogger('hcpsdk').addHandler(logging.NullHandler())

version = _Version()


class HcpsdkError(Exception):
    """
    Raised on generic errors in **hcpsdk**.
    """
    def __init__(self, reason):
        """
        :param reason:  an error description
        """
        self.args = (reason,)


class HcpsdkCantConnectError(HcpsdkError):
    """
    Raised if a connection couldn't be established.
    """
    def __init__(self, reason):
        """
        :param reason:  an error description
        """
        self.args = (reason,)


class HcpsdkTimeoutError(HcpsdkError):
    """
    Raised if a Connection timed out.
    """
    def __init__(self, reason):
        """
        :param reason:  an error description
        """
        self.args = (reason,)


class HcpsdkCertificateError(HcpsdkError):
    """
    Raised if the *SSL context* doesn't verify a certificate
    presented by HCP.
    """
    def __init__(self, reason):
        """
        :param reason:  an error description
        """
        self.args = (reason,)


class HcpsdkReplicaInitError(HcpsdkError):
    """
    Raised if the setup of the internal *Target* for the replica HCP failed
    (typically, this is a name resolution problem). **If
    this exception is raised, the primary Target's init failed, too.**
    You'll need to retry!
    """
    def __init__(self, reason):
        """
        :param reason:  an error description
        """
        self.args = (reason,)


class HcpsdkPortError(HcpsdkError):
    """
    Raised if the Target is initialized with an invalid port.
    """
    def __init__(self, reason):
        """
        :param reason:  an error description
        """
        self.args = (reason,)

# Port constants
P_HTTP = 80
P_HTTPS = 443
P_MGMT = 8000
P_SEARCH = 8888
P_MAPI = 9090


# Interface constants
I_DUMMY = 'I_DUMMY'
I_NATIVE = 'I_NATIVE'
I_HS3 = 'I_HS3'
I_HSWIFT = 'I_HSWIFT'

# Replication strategie constants
RS_READ_ALLOWED = 1  # allow to read from replica (always)
RS_READ_ON_FAILOVER = 2  # automatically read from replica when failed over
RS_WRITE_ALLOWED = 4  # allow to write to replica (always, A/A links only)
RS_WRITE_ON_FAILOVER = 8  # allow to write to replica when failed over

# The ports used for https
SSL_PORTS = [443, 8000, 9090]


class BaseAuthorization(object):
    """
    Represents the authorization for a *Target*.

    This is a base class for all other *Authorization* classes, not intended for
    direct usage, but to be sub-classed for specific protocols.
    """
    def __init__(self):
        """
        Calculate or acquire the authorization token (or whatever needed) to build the
        required authorization header(s).
        """
        self.headers = {}   # the headers that authorize a request

    def _createauthorization(self):
        """
        Do whatever is needed to create the authorization token.

        This method needs to be overwritten to suite the needs of a specific
        protocol.

        :return:    a dict holding the necessary headers
        """
        return self.headers

    def _refreshauthorization(self):
        """
        Do whatever is needed to refresh the authorization token.
        This method will be called by *Target* if an refresh of the authorization
        header(s) is required.

        This method needs to be overwritten if the specific protocol to refresh
        its authorization token from time to time.

        :return:    a dict holding the necessary headers
        :raises:    HcpsdkError
        """
        return self.headers

    def _getheaders(self):
        """
        This method will be called by *Target* to get the authorization header(s).

        :returns:   a dict holding the authorization headers
        :raises:    HcpsdkError, if no credentials are available
        """
        if self.headers:
            return self.headers
        else:
            raise HcpsdkError('Err: no authorization token available')


class DummyAuthorization(BaseAuthorization):
    """
    Dummy authorization for the :term:`Default Namespace <Default Namespace>`.
    """
    def __init__(self):
        super().__init__()
        self.logger = logging.getLogger(__name__ + '.DummyAuthorization')
        self.headers = {'HCPSDK_DUMMY': 'DUMMY'}
        self.logger.debug('*I_DUMMY* dummy authorization initialized')


class NativeAuthorization(BaseAuthorization):
    """
    Authorization for native http/REST access to HCP.
    """
    def __init__(self, user, password):
        """
        :param user:        the data access user
        :param password:    his password
        """
        super().__init__()
        self.logger = logging.getLogger(__name__ + '.NativeAuthorization')
        self.headers = self._createauthorization(user, password)
        self.logger.debug('*I_NATIVE* authorization initialized for user: {}'
                          .format(user))
        self.logger.debug(
            'pre version 6:     Cookie: {}'.format(self.headers['Cookie']))
        self.logger.debug('version 6+: Authorization: {}'.format(
            self.headers['Authorization']))

    def _createauthorization(self, user, password):
        """
        Build the authorization headers by calculation from user and password.

        :param user:        the name of a local HCP user
        :param password:    his password
        :return:            a dict holding the necessary headers
        """
        token = b64encode(user.encode()).decode() + ":" + md5(
            password.encode()).hexdigest()
        return {"Authorization": 'HCP {}'.format(token),
                "Cookie": "hcp-ns-auth={0}".format(token)}

class NativeADAuthorization(BaseAuthorization):
    """
    Authorization for native http/REST access to HCP using an Active Directory
    user.

    Supported with HCP 7.2.0 and later. The user needs to be a member of the
    Active Directory domain in which HCP is joined.
    """
    def __init__(self, user, password):
        """
        :param user:        an Active Directory user
        :param password:    his password
        """
        super().__init__()
        self.logger = logging.getLogger(__name__ + '.NativeADAuthorization')
        self.headers = self._createauthorization(user, password)
        self.logger.debug('*I_NATIVE* authorization initialized for user: {}'
                          .format(user))
        self.logger.debug('version 7.2+: Authorization: {}'.format(
            self.headers['Authorization']))

    def _createauthorization(self, user, password):
        """
        Build the authorization headers by calculation from user and password.

        :param user:        the name of a local HCP user
        :param password:    his password
        :return:            a dict holding the necessary headers
        """
        token = b64encode(user.encode()).decode() + ":" + md5(
            password.encode()).hexdigest()
        return {"Authorization": 'AD {}'.format(token)}

class LocalSwiftAuthorization(BaseAuthorization):
    """
    Authorization for local :term:`HSwift <HSwift>` access to
    HCP (w/o Keystone).
    """
    def __init__(self, user, password):
        """
        :param user:        the data access user
        :param password:    his password
        """
        super().__init__()
        self.logger = logging.getLogger(__name__ + '.LocalSwiftAuthorization')
        self.headers = self._createauthorization(user, password)
        self.logger.debug('Local HSwift: X-Auth-Token: {}'.format(
            self.headers['X-Auth-Token']))

    def _createauthorization(self, user, password):
        """
        Build the authorization headers by calculation from user and password.

        :param user:        the name of a local HCP user
        :param password:    his password
        :return:            a dict holding the necessary headers
        """
        token = b64encode(user.encode()).decode() + ":" + md5(
            password.encode()).hexdigest()
        return {"X-Auth-Token": "HCP {}".format(token)}


class Target(object):
    """
    This is the a central access point to an HCP target (and its replica,
    eventually). It caches the FQDN and the port and queries the provided
    *Authorization* object for the required authorization token.
    """

    def __init__(self, fqdn, authorization, port=443, dnscache=False,
                 sslcontext=SSL_NOVERIFY, interface=I_NATIVE,
                 replica_fqdn=None, replica_strategy=None):
        """
        :param fqdn:                ([namespace.]tenant.hcp.loc)
        :param authorization:       an instance of one of BaseAuthorization's subclasses
        :param port:                one of the port constants (*hcpsdk.P_**)
        :param dnscache:            if True, use the system resolver (which **might** do
                                    local caching), else use an internal resolver,
                                    bypassing any cache available
        :param sslcontext:          the context used to handle https requests; defaults to
                                    no certificate verification
        :param interface:           the HCP interface to use (I_NATIVE)
        :param replica_fqdn:        the replica HCP's FQDN
        :param replica_strategy:    OR'ed combination of the RS_* modes
        :raises:                    *ips.IpsError* if DNS query fails, *HcpsdkError* in all
                                    other fault cases
        """
        self.logger = logging.getLogger(__name__ + '.Target')
        self.__fqdn = fqdn
        self.__authorization = authorization
        self.__dnscache = dnscache
        self.__sslcontext = sslcontext
        self.__headers = {'Host': self.__fqdn}
        self.__port = port
        self.__ssl = self.__port in SSL_PORTS

        self.__interface = interface
        self.__replica = None  # placeholder for a replica's *Target* object
        self.__replica_strategy = replica_strategy

        # instantiate an IP address circler for this Target
        try:
            self.ipaddrqry = ips.Circle(self.__fqdn, port=self.__port,
                                        dnscache=self.__dnscache)
        except ips.IpsError as e:
            self.logger.debug(e, exc_info=True)
            raise ips.IpsError(e)
        except Exception as e:
            raise HcpsdkError(e)

        # noinspection PyProtectedMember
        self.logger.debug('Target initialized: {}:{} - SSL = {}'
                          .format(self.__fqdn, self.__port, self.__ssl))

        # If we have *replica_fqdn*, try to init its *Target* object
        if replica_fqdn:
            # try:
            #     self.__replica = Target(replica_fqdn, user, password,
            #                           self.__port, interface=self.__interface)
            # except HcpsdkError as e:
            #     raise HcpsdkReplicaInitError(e)
            raise HcpsdkReplicaInitError('Error: not yet implemented')

    def getaddr(self):
        """
        Convenience method to get an IP address out of the pool.

        :return:    an IP address (as string)
        """
        # noinspection PyProtectedMember
        return self.ipaddrqry._addr()

    # properties for the read-only attributes
    def __getfqdn(self):
        return self.__fqdn
    fqdn = property(__getfqdn, None, None,
                    'The FQDN for which this object was initialized (r/o)')

    def __getinterface(self):
        return self.__interface
    interface = property(__getinterface, None, None,
                         'The HCP interface used (r/o)')

    def __getport(self):
        return self.__port
    port = property(__getport, None, None,
                    'The target port in use (r/o)')

    def __getssl(self):
        return self.__ssl
    ssl = property(__getssl, None, None,
                    'Indicates if SSL is used (r/o)')

    def __getsslcontext(self):
        return self.__sslcontext
    sslcontext = property(__getsslcontext, None, None,
                    'The assigned SSL context (r/o)')

    def __getaddresses(self):
        return self.ipaddrqry._addresses
    addresses = property(__getaddresses, None, None,
                    'The list of resolved IP addresses for this target (r/o)')

    def __getheaders(self):
        tmp = self.__headers.copy()
        tmp.update(self.__authorization._getheaders())
        return tmp
    headers = property(__getheaders, None, None,
                    'The calculated authorization headers (r/o)')

    def __getreplica(self):
        return self.__replica
    replica = property(__getreplica, None, None,
                    'The target object for the HCP replica, if set (r/o)')

    def __getreplica_strategy(self):
        return self.__replica_strategy
    replica_strategy = property(__getreplica_strategy, None, None,
                    'The replica strategy selected (r/o)')

    def __repr__(self):
        return "<{} class at {}>".format(Target.__name__, id(self))

    def __str__(self):
        return "<{} class initialized for {}>".format(Target.__name__,
                                                      self.__fqdn)


class Connection(object):
    """
    This class represents a Connection to HCP,
    caching the related parameters.
    """

    # noinspection PyShadowingNames
    def __init__(self, target, timeout=30, idletime=30, retries=0,
                 debuglevel=0, sock_keepalive=False,
                 tcp_keepalive=60, tcp_keepintvl=60, tcp_keepcnt=3):
        """
        :param target:          an initialized Target object
        :param timeout:         the timeout for this Connection (secs)
        :param idletime:        the time the Connection shall stay persistence
                                when idle (secs)
        :param retries:         the number of retries until giving up on a
                                Request
        :param debuglevel:      0..9 -see->
                                `http.client.HTTPconnection <https://docs.python.org/3/library/http.client.html?highlight=http.client#http.client.HTTPConnection.set_debuglevel>`_
        :param sock_keepalive:  enable TCP keepalive, if True
        :param tcp_keepalive:   idle time used when SO_KEEPALIVE is enable
        :param tcp_keepintvl:   interval between keepalives
        :param tcp_keepcnt:     number of keepalives before close

        *Connection()* retries *request()s* if:
            a)  the underlying connection has been closed by HCP before
                *idletime* has passed (the request will be retried using the
                existing connection context) or
            b)  a timeout emerges during an active request, in which case the
                connection is closed, *Target()* is urged to refresh its cache
                of IP addresses, a fresh IP address is acquired from the cache
                and the connection is setup from scratch.

        You should rarely need this, but if you have a device in the data path
        that limits the time an idle connection can be open, this might be of
        help:

            Setting *sock_keepalive* to *True* enables TCP keep-alive for
            this connection. *tcp_keepalive* defines the idle time before a
            keep-alive packet is first sent, *tcp_keepintvl* is the time between
            keep-alive packets and *tcp_keepcnt* is the number of keep-alive
            packets to be sent before failing the connection in case the remote
            end doesn't answer.  See ``man tcp`` for the details.

            ..  versionadded:: 0.9.4.3
        """
        self.logger = logging.getLogger(__name__ + '.Connection')

        # This is here to allow test cases to inject error situations.
        # You need to set _fail to an exception object...
        self._fail = None
        #################

        self.__target = target  # an initialized Target object
        self.__address = None  # the assigned IP address to use
        self.__timeout = timeout  # the timeout for this Connection (secs)
        self.__idletime = float(idletime)  # the time the Connection shall stay open since last usage (secs)
        self.__debuglevel = debuglevel  # 0..9 -see-> http.client.HTTP[S]connetion
        self.__retries = retries  # the number of retries until giving up on a Request
        self.sock_keepalive = sock_keepalive
        self.tcp_keepalive = tcp_keepalive
        self.tcp_keepintvl = tcp_keepintvl
        self.tcp_keepcnt = tcp_keepcnt

        self.__sslcontext = self.__target.sslcontext
        self.__con = None  # http.client.HTTP[S]Connection object
        self._response = None

        self.__connect_time = 0.0  # record the time the connect() call took
        self.__service_time1 = 0.0  # the time a single step took (connect, 1st read, ...)
        self.__service_time2 = 0.0  # the time a Request took incl. all reads, but w/o connect

        self.idletimer = None  # used to hold a threading.Timer() object

        self.logger.log(logging.DEBUG,
                        'Connection object initialized: IP {} ({}) - timeout: '
                        '{} - idletime: {} - retries: {}'
                        .format(self.__address, self.__target.fqdn,
                                self.__timeout, self.__idletime,
                                self.__retries))
        if self.__sslcontext:
            self.logger.log(logging.DEBUG,
                            'SSLcontext = {}'.format(self.__sslcontext))

    def _set_idletimer(self):
        """
        Create and start a timer
        """
        self._cancel_idletimer()  # as a prevention, cancel a running timer
        self.idletimer = Timer(self.__idletime, self.__cancel_idletimer)
        self.idletimer.start()
        # self.logger.log(logging.DEBUG, 'idletimer started: {}'.format(self.idletimer))

    def _cancel_idletimer(self):
        """
        Cancel an active Connection keep-alive timer - manually called
        """
        if self.idletimer:
            self.idletimer.cancel()
            # self.logger.log(logging.DEBUG, 'idletimer canceled: {}'.format(self.idletimer))
            self.idletimer = None
        # else:
        #     self.logger.log(logging.DEBUG,
        #                     'tried to cancel a non-existing idletimer (pretty OK)'.format(self.idletimer))

    def __cancel_idletimer(self):
        """
        Cancel an active Connection keep-alive timer - auto-called if timer
        has passed
        """
        if self.idletimer:
            self.idletimer.cancel()
            self.close()
            self.logger.log(logging.DEBUG,
                            'idletimer timed out: {}'.format(self.idletimer))
            self.idletimer = None

    def _connect(self):
        """
        Open a new Connection and return the Connection object
        """
        self.__address = self.__target.getaddr()

        if self.__target.ssl:
            c_t = time.time()
            con = httpclient.HTTPSConnection(self.__address,
                                             port=self.__target.port,
                                             timeout=self.__timeout,
                                             context=self.__sslcontext,
                                             sock_keepalive=self.sock_keepalive,
                                             tcp_keepalive=self.tcp_keepalive,
                                             tcp_keepintvl=self.tcp_keepintvl,
                                             tcp_keepcnt=self.tcp_keepcnt)
            self.__connect_time = time.time() - c_t
        else:
            c_t = time.time()
            con = httpclient.HTTPConnection(self.__address,
                                            port=self.__target.port,
                                            timeout=self.__timeout,
                                            sock_keepalive=self.sock_keepalive,
                                            tcp_keepalive=self.tcp_keepalive,
                                            tcp_keepintvl=self.tcp_keepintvl,
                                            tcp_keepcnt=self.tcp_keepcnt)
            self.__connect_time = time.time() - c_t
        self.logger.log(logging.DEBUG,
                        'Connection open: IP {} ({}) - connect_time: {:0.17f}'
                        .format(self.__address, self.__target.fqdn,
                                self.__connect_time))

        if self.__debuglevel:
            con.set_debuglevel(self.__debuglevel)
        return con

    def request(self, method, url, body=None, params=None, headers=None):
        """
        Wraps the *http.client.HTTP[s]Connection.Request()* method to be able to
        catch any exception that might happen plus to be able to trigger
        hcpsdk.Target to do a new DNS query.

        *Url* and *params* will be urlencoded, by default.

        **Beside of *method*, all arguments are valid for the convenience methods, too.**

        :param method:  any valid http method (GET,HEAD,PUT,POST,DELETE)
        :param url:     the url to access w/o the server part (i.e: /rest/path/object)
        :param body:    the payload to send (see *http.client* documentation
                        for details)
        :param params:  a dictionary with parameters to be added to the Request:

                        ``{'verbose': 'true', 'retention': 'A+10y', ...}``

                        or a list of 2-tuples:

                        ``[('verbose', 'true'), ('retention', 'A+10y'), ...]``

        :param headers: a dictionary holding additional key/value pairs to add to the
                        auto-prepared header
        :return:        the original *Response* object received from
                        *http.client.HTTP[S]Connection.requests()*.
        :raises:        one of the *hcpsdk.Hcpsdk[..]Error*\ s or
                        *hcpsdk.ips.IpsError* in case an IP address cache refresh failed
        """
        self._cancel_idletimer()  # 1st, cancel the idletimer
        if not headers:
            headers = self.__target.headers
        else:
            headers.update(self.__target.headers)

        # make sure that the URL and params are proper encoded
        url = quote(url, safe='/')
        if params:
            url = url + '?' + urlencode(params)
        self.logger.log(logging.DEBUG, 'URL = {}'.format(url))

        initialretry = False    # used if connection isn't open
        retryonfailure = False  # used for retries on failures
        retries = 0             # - " -
        while True:
            try:
                if retryonfailure:
                    retryonfailure = False
                    self.close()
                    self.__target.ipaddrqry.refresh()
                    self.__con = self._connect()
                if initialretry:
                    self.close()
                    self.__con = self._connect()
                    initialretry = False

                # This is to allow a test case to inject an error situation...
                if self._fail:
                    __e = self._fail
                    self._fail = None
                    raise __e('test case')
                ####################

                self.logger.log(logging.DEBUG, '{}: About to request for {}'
                                .format(method, url))
                s_t = time.time()
                self.__con.request(method, url, body=body, headers=headers)
            except ips.IpsError as e:
                # This is a trigger for the case that *hcpsdk.ips* isn't able
                # to resolve IP addresses - we simple forward it, as we can't
                # resolve.
                self._fail = None
                raise
            except ConnectionRefusedError as e:
                # This is a trigger for the case that we were able to get an
                # IP address, but a connection to it was actively refused.
                self.close()
                raise HcpsdkError('Unable to connect ({})'
                                  .format(str(e)))
            except (http.client.NotConnected, AttributeError) as e:
                # This is a trigger for the case the Connection is not open
                # (not yet opened or has been closed by being not used for some
                # time). So, we open up a new Connection and start over by
                # calling our self again...
                self._fail = None
                if not initialretry:
                    self.logger.log(logging.DEBUG,
                                    'Connection needs to be opened')
                    initialretry = True
                    continue
                else:
                    self.close()
                    raise HcpsdkError('Can\'t connect, retry failed ({})'
                                      .format(str(e)))
            except ConnectionAbortedError as e:
                # This is a trigger for the case that HCP aborts a connection
                # for whatever reason. It also serves to catch WinError 10053
                # on Windows, which stands for a close caused by the OS. We
                # close the connection, force the target to refresh its address
                # list and retry with a new connection.
                self._fail = None
                self.logger.debug(
                    'ConnectionAbortedError: {} Request for {} failed ({})'
                    .format(method, url, e))
                if retries < self.__retries:
                    retries += 1
                    retryonfailure = True
                    self.logger.log(logging.DEBUG,
                                    'ConnectionAbortedError - retry # {}'
                                    .format(retries))
                    continue
                else:
                    self.logger.log(logging.DEBUG,
                                    'ConnectionAbortedError ({} retries), '
                                    'giving up'.format(retries))
                    self.close()
                    raise HcpsdkTimeoutError(
                        'ConnectionAbortedError (giving up after {} retries) '
                        '- {}'.format(retries, url))
            except http.client.CannotSendRequest as e:
                # If this gets raised, the underlying connection seems to be in
                # a state where it can't handle a new request, yet. We'll try
                # it with the same approach as with the ConnectionAbortedError
                self._fail = None
                self.logger.log(logging.DEBUG,
                                'CannotSendRequest: {} Request for {} failed '
                                '({})'.format(method, url, e))
                if retries < self.__retries:
                    retries += 1
                    initialretry = True
                    self.logger.log(logging.DEBUG,
                                    'CannotSendRequest - retry # {}'.format(
                                        retries))
                    continue
                else:
                    self.logger.log(logging.DEBUG,
                                    'CannotSendRequest ({} retries), giving up'
                                    .format(retries))
                    self.close()
                    raise HcpsdkTimeoutError(
                        'CannotSendRequest (giving up after {} retries) - {}'
                            .format(retries, url))
            except http.client.ResponseNotReady as e:
                # If this gets raised, the underlying connection seems to be in
                # a state where it can't handle a new request, yet. We'll try it
                # with the same approach as with the ConnectionAbortedError...
                self._fail = None
                self.logger.debug(
                    'http.client.ResponseNotReady: {} Request for {} failed '
                    '({})'.format(method, url, e))
                if retries < self.__retries:
                    retries += 1
                    retryonfailure = True
                    self.logger.log(logging.DEBUG,
                                    'http.client.ResponseNotReady - retry # {}'
                                    .format(retries))
                    continue
                else:
                    self.logger.log(logging.DEBUG,
                                    'http.client.ResponseNotReady ({} retries)'
                                    ', giving up'.format(retries))
                    self.close()
                    raise HcpsdkTimeoutError(
                        'http.client.ResponseNotReady (giving up after {} '
                        'retries) - {}'.format(retries, url))

            except ssl.SSLError as e:
                # This is a blocking issue - will *not* retry and will close
                # the underlying connection.
                self._fail = None
                self.logger.log(logging.DEBUG, 'ssl.SSLError: {}'
                                .format(str(e)))
                self.close()
                raise HcpsdkCertificateError(str(e))
            except (TimeoutError, socket.timeout, BrokenPipeError) as e:
                # We will retry in this case (if retries have been asked for).
                # If we fail we close the underlying connection.
                self._fail = None
                self.logger.debug('TimeoutError: {} Request for {} failed ({})'
                                  .format(method, url, e))
                if retries < self.__retries:
                    retries += 1
                    retryonfailure = True
                    self.logger.log(logging.DEBUG, 'TimeoutError - retry # {}'
                                    .format(retries))
                    continue
                else:
                    self.logger.log(logging.DEBUG, 'TimeoutError ({} retries),'
                                                   ' giving up'
                                    .format(retries))
                    self.close()
                    raise HcpsdkTimeoutError('Timeout ({} retries) - {}'
                                             .format(retries, url))
            except http.client.HTTPException as e:
                # Again, there might be no recovery from this, so we close the
                # underlying connection and give up.
                self._fail = None
                self.logger.exception('unexpected HTTPException')
                self.close()
                raise HcpsdkError(str(e))
            except Exception as e:
                # Again, there might be no recovery from this, so we close the
                # underlying connection and give up.
                self._fail = None
                self.logger.exception('unexpected Exception')
                self.close()
                raise HcpsdkError(str(e))
            else:
                self.__service_time1 = self.__service_time2 = time.time() - s_t
                self.logger.log(logging.DEBUG,
                                '{} Request for {} - service_time1&2 = '
                                '{:0.17f}'
                                .format(method, url, self.__service_time1))

                try:
                    self._response = self.__con.getresponse()
                except (TimeoutError, socket.timeout, BrokenPipeError) as e:
                    if retries < self.__retries:
                        retries += 1
                        self.logger.log(logging.DEBUG,
                                        'TimeoutError while getting response '
                                        '- retry # {}'.format(retries))
                        continue
                    else:
                        self.logger.log(logging.DEBUG,
                                        'TimeoutError while getting response '
                                        '({} retries), giving up'
                                        .format(retries))
                        self.close()
                        raise HcpsdkTimeoutError(
                            'Timeout ({} retries) - {}'.format(retries, url))
                except (OSError, http.client.BadStatusLine) as e:
                    # BadStatusLine most likely means that HCP has closed the connection.
                    # Same for OSError 9
                    # ('HTTP Persistent Connection Timeout Interval' < Connection.timeout)
                    # So, we close the connection here and trigger a retry...
                    self.close()
                    if retries < self.__retries:
                        retries += 1
                        retryonfailure = True
                        self.logger.log(logging.DEBUG,
                                        'HCP most likely closed the connection'
                                        ' ({}) - retry # {}'
                                        .format(str(e), retries))
                        continue
                    else:
                        self.logger.log(logging.DEBUG,
                                        'HCP most likely closed the connection'
                                        ' ({} - {} retries, giving up)'
                                        .format(str(e), retries))
                        raise HcpsdkTimeoutError(
                            'HCP most likely closed the connection ({} - {} '
                            'retries) - {}'
                            .format(str(e), retries, url))
                except http.client.ResponseNotReady as e:
                    # If this gets raised, the underlying connection seems to
                    # be in a state where it can't handle a new request, yet.
                    # We'll try it with the same approach as with the
                    # ConnectionAbortedError...
                    self._fail = None
                    self.logger.debug(
                        'http.client.ResponseNotReady: {} getresponse() for '
                        '{} failed ({})'.format(method, url, e))
                    if retries < self.__retries:
                        retries += 1
                        retryonfailure = True
                        self.logger.log(logging.DEBUG,
                                        'http.client.ResponseNotReady - retry '
                                        '# {}'.format(retries))
                        continue
                    else:
                        self.logger.log(logging.DEBUG,
                                        'http.client.ResponseNotReady ({} '
                                        'retries), giving up'
                                        .format(retries))
                        self.close()
                        raise HcpsdkTimeoutError(
                            'http.client.ResponseNotReady (giving up after {}'
                            ' retries) - {}'
                            .format(retries, url))
                except Exception as e:
                    self.logger.exception(
                        'Exception not catched in hcpsdk.__init__: {}'.format(
                            str(e)))
                else:
                    self.__service_time2 = time.time() - s_t
                    self.logger.log(logging.DEBUG,
                                    '{} Request for {} - after getResponse(): '
                                    'service_time2 = {:0.17f}'
                                    .format(method, url, self.__service_time2))

            self._set_idletimer()
            return self._response

    def getheader(self, *args, **kwargs):
        """
        Used to get a single *Response* header. Wraps
        *http.client.Response.getheader()*. Arguments are simply passed
        through.
        """
        return self._response.getheader(*args, **kwargs)

    def getheaders(self):
        """
        Used to get a the *Response* headers. Wraps
        *http.client.Response.getheaders()*.
        """
        return self._response.getheaders()

    # noinspection PyUnusedLocal,PyPep8Naming
    def PUT(self, url, body=None, params=None, headers=None):
        """
        Convenience method for Request() - PUT an object.
        Cleans up and leaves the Connection ready for the next Request.
        For parameter description see *Request()*.
        """
        r = self.request('PUT', url, body, params, headers)
        r.read()  # clean up
        return r

    # noinspection PyPep8Naming
    def GET(self, url, params=None, headers=None):
        """
        Convenience method for Request() - GET an object.
        You need to fully *.read()* the requested content from the Connection
        before it can be used for another Request.
        For parameter description see *Request()*.
        """
        return self.request('GET', url, params=params, headers=headers)

    def HEAD(self, url, params=None, headers=None):
        """
        Convenience method for Request() - HEAD - get metadata of an object.
        Cleans up and leaves the Connection ready for the next Request.
        For parameter description see *Request()*.
        """
        r = self.request('HEAD', url, params=params, headers=headers)
        r.read()
        return r

    def POST(self, url, body=None, params=None, headers=None):
        """
        Convenience method for Request() - POST metadata.
        Does no clean-up, as a POST can have a response body!
        For parameter description see *Request()*.
        """
        return self.request('POST', url, body=body, params=params,
                            headers=headers)

    def DELETE(self, url, params=None, headers=None):
        """
        Convenience method for Request() - DELETE an object.
        Cleans up and leaves the Connection ready for the next Request.
        For parameter description see *Request()*.
        """
        r = self.request('DELETE', url, params=params, headers=headers)
        r.read()  # clean up
        return r

    def read(self, amt=None):
        """
        Read amt # of bytes (or all, if amt isn't given) from a *Response*.

        :param amt: number of bytes to read
        :return:    the requested number of bytes; fewer (or zero) bytes signal
                    end of transfer, which means that the Connection is ready
                    for another Request.
        :raises:    *HcpsdkTimeoutError* in case a socket.timeout was catched,
                    *HcpsdkError* in all other cases.
        """
        s_t = time.time()
        try:
            buf = self._response.read(amt)
            self.__service_time1 = time.time() - s_t
        except AttributeError as e:
            msg = 'faulty read: {}'.format(str(e))
            self.logger.log(logging.DEBUG, msg)
            raise HcpsdkError(msg)
        except socket.timeout as e:
            msg = 'read: {}'.format(str(e))
            self.logger.log(logging.DEBUG, msg)
            raise HcpsdkTimeoutError(msg)
        except (http.client.IncompleteRead, OSError) as e:
            msg = 'read error: {}'.format(str(e))
            self.logger.log(logging.DEBUG, msg)
            raise HcpsdkError(msg)
        else:
            self.__service_time2 += self.__service_time1
            readsize = len(buf)
            if readsize:
                self.logger.log(logging.DEBUG,
                                '(partial?) read {} bytes: service_time1/2 = '
                                '{:0.17f}/{:0.17f} secs'
                                .format(readsize, self.__service_time1,
                                        self.__service_time2))
            else:
                self.logger.log(logging.DEBUG,
                                'final read: service_time1/2 = {:0.17f}/'
                                '{:0.17f} secs'
                                .format(self.__service_time1,
                                        self.__service_time2))
            return buf

    def close(self):
        """
        Close the Connection.

        .. Warning::
           **It is essential to close the Connection**, as open connections
           might keep the program from terminating for at max *timeout*
           seconds, due to the fact that the timer used to keep the Connection
           persistent runs in a separate thread, which will be canceled on
           *close()*.
        """
        # noinspection PyBroadException
        if self.__con:
            try:
                self._cancel_idletimer()
                self.__con.close()
                self.__con = None
                self.logger.log(logging.DEBUG,
                                'Connection object closed: IP {} ({})'
                                .format(self.__address, self.__target.fqdn))
            except Exception as e:
                self.logger.exception('Connection object close failed: '
                                      'IP {} ({})'
                                .format(self.__address, self.__target.fqdn))

    # properties for externally visible attributes
    def __getaddress(self):
        return self.__address
    address = property(__getaddress, None, None,
                    'The IP address for which this object was initialized '
                    '(r/o)')

    def __getcon(self):
        return self.__con
    con = property(__getcon, None, None,
                    'The internal connection object (r/o)')

    def __getresponse(self):
        return self._response
    Response = property(__getresponse, None, None,
                        '.. deprecated:: 0.9.4.2\n\n'
                        '   Use **response** instead!')
    response = property(__getresponse, None, None,
                        'Exposition of the http.client.Response object for '
                        'the last Request (r/o)\n\n'
                        '.. versionadded:: 0.9.4.2')


    def __getresponse_status(self):
        return self._response.status
    response_status = property(__getresponse_status, None, None,
                               'The HTTP status code of the last Request '
                               '(r/o)')

    def __getresponse_reason(self):
        return self._response.reason
    response_reason = property(__getresponse_reason, None, None,
                               'The corresponding HTTP status message (r/o)')

    def __getconnect_time(self):
        if self.__connect_time > 0.0:
            return self.__connect_time
        else:
            return 0.00000000001
    connect_time = property(__getconnect_time, None, None,
                            'The time in seconds the last connect took (r/o)')

    def __getservice_time1(self):
        if self.__service_time1 > 0.0:
            return self.__service_time1
        else:
            return 0.00000000001
    service_time1 = property(__getservice_time1, None, None,
                             'The time in seconds the last action on a Request'
                             ' took. This can be the initial part of PUT/GET/'
                             'etc., or a single (possibly incomplete) read '
                             'from a Response (r/o)')

    def __getservice_time2(self):
        if self.__service_time2 > 0.0:
            return self.__service_time2
        else:
            return 0.00000000001
    service_time2 = property(__getservice_time2, None, None,
                             'Duration in secods of the complete Request up '
                             'to now. Sum of all ``service_time1`` during '
                             'handling a Request (r/o)')

    def __getdebug_level(self):
        return self.__debuglevel
    def __setdebug_level(self, value):
        if type(value) != int or value not in range(0,10):
            raise ValueError('debug_level must be in range 0..9')
        self.__debuglevel = value
        if self.__con:
            self.__con.set_debuglevel(self.__debuglevel)
        self.logger.debug('debug_level set to {}'.format(self.__debuglevel))
    debug_level = property(__getdebug_level, __setdebug_level,
                           'The debug level used by underlying '
                           '*http.client.HTTP(S)Connection* object (r/w)')

    def __repr__(self):
        return "<{} class at {}>".format(Connection.__name__, id(self))

    def __str__(self):
        return ("<{} class initialized for fqdn {} @ {}>"
                .format(Connection.__name__, self.__target.fqdn,
                        self.__address))


# helper functions
def checkport(target, port):
    """
    Check if an *hcpsdk.Target()* object is initialized with the correct port.

    :param target:  the *hcpsdk.Target()* object to check
    :param port:    the needed port
    :returns:       nothing
    :raises:        *hcpsdk.HcpsdkPortError* in case the port is invalid
    """
    logger = logging.getLogger(__name__)

    if target.port != port:
        raise HcpsdkPortError('Target initialized for port {}, not {}'
                              .format(target.port, port))

